# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various convolutional networks.
"""
from typing import List, Optional, Tuple, Union, Dict, Any

import torch

from texar.torch.core.layers import get_pooling_layer_hparams
from texar.torch.hyperparams import HParams
from texar.torch.modules.networks.network_base import FeedForwardNetworkBase
from texar.torch.utils.shapes import mask_sequences
from texar.torch.utils.utils import uniquify_str

__all__ = [
    "_to_list",
    "Conv1DNetwork",
]


def _to_list(value: Union[Dict[str, Any], List, Tuple, int], name=None,
             list_length=None):
    r"""Converts `hparams` value into a list.

    If :attr:`list_length` is given, then the canonicalized :attr:`value`
    must be of length :attr:`list_length`.
    """
    if not isinstance(value, (list, tuple)):
        if list_length is not None:
            value = [value] * list_length
        else:
            value = [value]
    if list_length is not None and len(value) != list_length:
        name = '' if name is None else name
        raise ValueError("hparams '%s' must be a list of length %d"
                         % (name, list_length))
    return value


class Conv1DNetwork(FeedForwardNetworkBase):
    r"""Simple `Conv-1D` network which consists of a sequence of convolutional
    layers followed with a sequence of dense layers.

    Args:
        in_channels (int): Number of channels in the input tensor.
        in_features (int): Size of the feature dimension in the input tensor.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.

    See :meth:`forward` for the inputs and outputs. If :attr:`"data_format"` is
    set to ``"channels_first"`` (this is the default), inputs must be a tensor
    of shape `[batch_size, channels, length]`. If :attr:`"data_format"` is set
    to ``"channels_last"``, inputs must be a tensor of shape
    `[batch_size, length, channels]`. For example, for sequence classification,
    `length` corresponds to time steps, and `channels` corresponds to embedding
    dim.

    Example:

    .. code-block:: python

        nn = Conv1DNetwork(in_channels=20, in_features=256) # Use the default

        inputs = torch.randn([64, 20, 256])
        outputs = nn(inputs)
        # outputs == Tensor of shape [64, 256], because the final dense layer
        # has size 256.

    .. document private functions
    """

    def __init__(self, in_channels: int, in_features: Optional[int] = None,
                 hparams=None):
        super().__init__(hparams=hparams)
        if self.hparams.num_dense_layers > 0 and in_features is None:
            raise ValueError("\"in_features\" cannot be None "
                             "if \"num_dense_layers\" > 0")

        # construct only non-dense layers first
        layer_hparams = self._build_non_dense_layer_hparams(
            in_channels=in_channels)
        self._build_layers(layers=None, layer_hparams=layer_hparams)
        if self.hparams.num_dense_layers > 0:
            if in_features is None:
                raise ValueError("\"in_features\" cannot be None "
                                 "if \"num_dense_layers\" > 0")
            ones = torch.ones(1, in_channels, in_features)
            input_size = self._infer_dense_layer_input_size(ones)
            layer_hparams = self._build_dense_hparams(
                in_features=input_size[1], layer_hparams=layer_hparams)
            self._build_layers(layers=None, layer_hparams=layer_hparams)

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.

        .. code-block:: python

            {
                # (1) Conv layers
                "num_conv_layers": 1,
                "out_channels": 128,
                "kernel_size": [3, 4, 5],
                "conv_activation": "ReLU",
                "conv_activation_kwargs": None,
                "other_conv_kwargs": {},
                "data_format": "channels_first",
                # (2) Pooling layers
                "pooling": "MaxPool1d",
                "pool_size": None,
                "pool_stride": 1,
                "other_pool_kwargs": {},
                # (3) Dense layers
                "num_dense_layers": 1,
                "out_features": 256,
                "dense_activation": None,
                "dense_activation_kwargs": None,
                "final_dense_activation": None,
                "final_dense_activation_kwargs": None,
                "other_dense_kwargs": None,
                # (4) Dropout
                "dropout_conv": [1],
                "dropout_dense": [],
                "dropout_rate": 0.75,
                # (5) Others
                "name": "conv1d_network"
            }

        Here:

        1. For **convolutional** layers:

           `"num_conv_layers"`: int
               Number of convolutional layers.

           `"out_channels"`: int or list
               The number of out_channels in the convolution, i.e., the
               dimensionality of the output space.

               - If ``"num_conv_layers"`` > 1 and ``"out_channels"`` is an int,
                 all convolution layers will have the same number of output
                 channels.

               - If ``"num_conv_layers"`` > 1 and ``"out_channels"`` is a list,
                 the length must equal ``"num_conv_layers"``. The number of
                 output channels of each convolution layer will be the
                 corresponding element from this list.

           `"kernel_size"`: int or list
               Lengths of 1D convolution windows.

               - If `"num_conv_layers"` = 1, this can also be a ``int`` list of
                 arbitrary length denoting differently sized convolution
                 windows. The number of output channels of each size is
                 specified by ``"out_channels"``.
                 For example, the default values will create 3 convolution
                 layers, each of which has kernel size of 3, 4, and 5,
                 respectively, and has output channel 128.

               - If `"num_conv_layers"` > 1, this must be a list of length
                 ``"num_conv_layers"``. Each element can be an ``int`` or a
                 ``int`` list of arbitrary length denoting the kernel size of
                 each layer.

           `"conv_activation"`: str or callable
               Activation applied to the output of the convolutional
               layers. Set to `None` to maintain a linear activation.
               See :func:`~texar.torch.core.get_layer` for more details.

           `"conv_activation_kwargs"`: dict, optional
               Keyword arguments for the activation following the convolutional
               layer. See :func:`~texar.torch.core.get_layer` for more details.

           `"other_conv_kwargs"`: list or dict, optional
               Other keyword arguments for :torch_nn:`Conv1d` constructor,
               e.g., ``padding``.

               - If a dict, the same dict is applied to all the convolution
                 layers.

               - If a list, the length must equal ``"num_conv_layers"``. This
                 list can contain nested lists. If the convolution layer at
                 index i has multiple kernel sizes, then the corresponding
                 element of this list can also be a list of length equal to
                 ``"kernel_size"`` at index i. If the element at index i is
                 instead a dict, then the same dict gets applied to all the
                 convolution layers at index i.

           `"data_format"`: str, optional
               Data format of the input tensor. Defaults to ``channels_first``
               denoting the first dimension to be the channel dimension. Set it
               to ``channels_last`` to treat last dimension as the channel
               dimension. This argument can also be passed in ``forward``
               function, in which case the value specified here will be
               ignored.

        2. For **pooling** layers:

           `"pooling"`: str or class or instance
               Pooling layer after each of the convolutional layer(s). Can be a
               pooling layer class, its name or module path, or a class
               instance.

           `"pool_size"`: int or list, optional
               Size of the pooling window. If an ``int``, all pooling layer
               will have the same pool size. If a list, the list length must
               equal ``"num_conv_layers"``. If `None` and the pooling type
               is either :torch_docs:`MaxPool1d <nn.html#maxpool1d>` or
               :torch_docs:`AvgPool1d <nn.html#avgpool1d>`, the pool size will
               be set to input size. That is, the output of the pooling layer
               is a single unit.

           `"pool_stride"`: int or list, optional
               Strides of the pooling operation. If an ``int``, all
               layers will have the same stride. If a list, the list length
               must equal ``"num_conv_layers"``.

           `"other_pool_kwargs"`: list or dict, optional
               Other keyword arguments for pooling layer class constructor.

               - If a dict, the same dict is applied to all the pooling layers.

               - If a list, the length must equal ``"num_conv_layers"``. The
                 pooling arguments for layer i will be the element at index i
                 from this list.

        3. For **dense** layers (note that here dense layers always follow
           convolutional and pooling layers):

           `"num_dense_layers"`: int
               Number of dense layers.

           `"out_features"`: int or list
               Dimension of features after the dense layers. If an
               ``int``, all dense layers will have the same feature dimension.
               If a list of ``int``, the list length must equal
               ``"num_dense_layers"``.

           `"dense_activation"`: str or callable
               Activation function applied to the output of the dense
               layers **except** the last dense layer output. Set to
               `None` to maintain a linear activation.

           `"dense_activation_kwargs"`: dict, optional
               Keyword arguments for dense layer activation functions before
               the last dense layer.

           `"final_dense_activation"`: str or callable
               Activation function applied to the output of the **last** dense
               layer. Set to `None` to maintain a linear activation.

           `"final_dense_activation_kwargs"`: dict, optional
               Keyword arguments for the activation function of last
               dense layer.

           `"other_dense_kwargs"`: dict, optional
               Other keyword arguments for dense layer class constructor.

        4. For **dropouts**:

           `"dropout_conv"`: int or list
               The indices of convolutional layers (starting from 0) whose
               **inputs** are applied with dropout.
               The index = :attr:`num_conv_layers` means dropout applies to the
               final convolutional layer output. For example,

               .. code-block:: python

                   {
                       "num_conv_layers": 2,
                       "dropout_conv": [0, 2]
                   }

               will leads to a series of layers as
               `-dropout-conv0-conv1-dropout-`.

               The dropout mode (training or not) is controlled
               by :attr:`self.training`.

           `"dropout_dense"`: int or list
               Same as ``"dropout_conv"`` but applied to dense layers (index
               starting from 0).

           `"dropout_rate"`: float
               The dropout rate, between 0 and 1. For example,
               ``"dropout_rate": 0.1`` would drop out 10% of elements.

        5. Others:

           `"name"`: str
               Name of the network.
        """
        return {
            # (1) Conv layers
            "num_conv_layers": 1,
            "out_channels": 128,
            "kernel_size": [3, 4, 5],
            "conv_activation": "ReLU",
            "conv_activation_kwargs": None,
            "other_conv_kwargs": {},
            "data_format": "channels_first",
            # (2) Pooling layers
            "pooling": "MaxPool1d",
            "pool_size": None,
            "pool_stride": 1,
            "other_pool_kwargs": {},
            # (3) Dense layers
            "num_dense_layers": 1,
            "out_features": 256,
            "dense_activation": None,
            "dense_activation_kwargs": None,
            "final_dense_activation": None,
            "final_dense_activation_kwargs": None,
            "other_dense_kwargs": None,
            # (4) Dropout
            "dropout_conv": [1],
            "dropout_dense": [],
            "dropout_rate": 0.75,
            # (5) Others
            "name": "conv1d_network",
            "@no_typecheck": ["out_channels", "kernel_size", "conv_activation",
                              "other_conv_kwargs", "pool_size", "pool_stride",
                              "other_pool_kwargs", "out_features",
                              "dense_activation", "dropout_conv",
                              "dropout_dense"]
        }

    def _build_pool_hparams(self):
        pool_type = self._hparams.pooling
        if pool_type == "MaxPool":
            pool_type = "MaxPool1d"
        elif pool_type == "AvgPool":
            pool_type = "AvgPool1d"

        npool = self._hparams.num_conv_layers
        kernel_size = _to_list(self._hparams.pool_size, "pool_size", npool)
        stride = _to_list(self._hparams.pool_stride, "pool_stride", npool)

        other_kwargs = self._hparams.other_pool_kwargs
        if isinstance(other_kwargs, HParams):
            other_kwargs = other_kwargs.todict()
            other_kwargs = _to_list(other_kwargs, "other_kwargs", npool)
        elif isinstance(other_kwargs, (list, tuple)):
            if len(other_kwargs) != npool:
                raise ValueError("The length of hparams['other_pool_kwargs'] "
                                 "must equal 'num_conv_layers'")
        else:
            raise ValueError("hparams['other_pool_kwargs'] must be either a "
                             "dict or list/tuple")

        pool_hparams = []
        for i in range(npool):
            kwargs_i = {"kernel_size": kernel_size[i], "stride": stride[i]}
            kwargs_i.update(other_kwargs[i])
            pool_hparams_ = get_pooling_layer_hparams({"type": pool_type,
                                                       "kwargs": kwargs_i})
            pool_hparams.append(pool_hparams_)

        return pool_hparams

    def _build_conv1d_hparams(self, in_channels, pool_hparams):
        r"""Creates the hparams for each of the convolutional layers usable for
        :func:`texar.torch.core.layers.get_layer`.
        """
        nconv = self._hparams.num_conv_layers
        if len(pool_hparams) != nconv:
            raise ValueError("`pool_hparams` must be of length %d" % nconv)

        in_channels = [in_channels]
        out_channels = _to_list(self._hparams.out_channels, 'out_channels',
                                nconv)
        # because in_channels(i) = out_channels(i-1)
        in_channels.extend(out_channels[:-1])

        if nconv == 1:
            kernel_size = _to_list(self._hparams.kernel_size)
            if not isinstance(kernel_size[0], (list, tuple)):
                kernel_size = [kernel_size]
        elif nconv > 1:
            kernel_size = _to_list(self._hparams.kernel_size,
                                   'kernel_size', nconv)
            kernel_size = [_to_list(ks) for ks in kernel_size]

        other_kwargs = self._hparams.other_conv_kwargs
        if isinstance(other_kwargs, HParams):
            other_kwargs = other_kwargs.todict()
            other_kwargs = _to_list(other_kwargs, "other_conv_kwargs", nconv)
        elif isinstance(other_kwargs, (list, tuple)):
            if len(other_kwargs) != nconv:
                raise ValueError("The length of hparams['other_conv_kwargs'] "
                                 "must be equal to 'num_conv_layers'")
        else:
            raise ValueError("hparams['other_conv_kwargs'] must be a either "
                             "a dict or a list.")

        def _activation_hparams(name, kwargs=None):
            if kwargs is not None:
                return {"type": name, "kwargs": kwargs}
            else:
                return {"type": name, "kwargs": {}}

        conv_pool_hparams = []
        for i in range(nconv):
            hparams_i = []
            names = []
            if isinstance(other_kwargs[i], dict):
                other_kwargs[i] = _to_list(other_kwargs[i], "other_kwargs[i]",
                                           len(kernel_size[i]))
            elif (isinstance(other_kwargs[i], (list, tuple))
                  and len(other_kwargs[i]) != len(kernel_size[i])):
                raise ValueError("The length of hparams['other_conv_kwargs'][i]"
                                 " must be equal to the length of "
                                 "hparams['kernel_size'][i]")
            for idx, ks_ij in enumerate(kernel_size[i]):
                name = uniquify_str("conv_%d" % (i + 1), names)
                names.append(name)
                conv_kwargs_ij = {
                    "in_channels": in_channels[i],
                    "out_channels": out_channels[i],
                    "kernel_size": ks_ij
                }
                conv_kwargs_ij.update(other_kwargs[i][idx])
                hparams_i.append(
                    {"type": "Conv1d", "kwargs": conv_kwargs_ij})
            if len(hparams_i) == 1:
                if self._hparams.conv_activation:
                    layers = {
                        "layers": [hparams_i[0],
                                   _activation_hparams(
                                       self._hparams.conv_activation,
                                       self._hparams.conv_activation_kwargs)]}
                    sequential_layer = {"type": "Sequential", "kwargs": layers}
                    conv_pool_hparams.append([sequential_layer,
                                              pool_hparams[i]])
                else:
                    conv_pool_hparams.append([hparams_i[0], pool_hparams[i]])
            else:  # creates MergeLayer
                mrg_kwargs_layers = []
                for hparams_ij in hparams_i:
                    if self._hparams.conv_activation:
                        seq_kwargs_j = {
                            "layers": [
                                hparams_ij,
                                _activation_hparams(
                                    self._hparams.conv_activation,
                                    self._hparams.conv_activation_kwargs),
                                pool_hparams[i]
                            ]
                        }
                    else:
                        seq_kwargs_j = {"layers": [hparams_ij, pool_hparams[i]]}
                    mrg_kwargs_layers.append(
                        {"type": "Sequential", "kwargs": seq_kwargs_j})
                mrg_hparams = {"type": "MergeLayer",
                               "kwargs": {"layers": mrg_kwargs_layers}}
                conv_pool_hparams.append(mrg_hparams)

        return conv_pool_hparams

    def _build_dense_hparams(self, in_features: int, layer_hparams):
        ndense = self._hparams.num_dense_layers
        in_features = [in_features]
        out_features = _to_list(self._hparams.out_features, 'out_features',
                                ndense)
        # because in_features(i) = out_features(i-1)
        in_features.extend(out_features[:-1])
        other_kwargs = self._hparams.other_dense_kwargs or {}
        if isinstance(other_kwargs, HParams):
            other_kwargs = other_kwargs.todict()
        if not isinstance(other_kwargs, dict):
            raise ValueError("hparams['other_dense_kwargs'] must be a dict.")

        def _activation_hparams(name, kwargs=None):
            if kwargs is not None:
                return {"type": name, "kwargs": kwargs}
            else:
                return {"type": name, "kwargs": {}}

        dense_hparams = []
        for i in range(ndense):
            kwargs_i = {"in_features": in_features[i],
                        "out_features": out_features[i]}
            kwargs_i.update(other_kwargs)

            dense_hparams_i = {"type": "Linear", "kwargs": kwargs_i}
            if i < ndense - 1 and self._hparams.dense_activation is not None:
                layers = {
                    "layers": [dense_hparams_i,
                               _activation_hparams(
                                   self._hparams.dense_activation,
                                   self._hparams.dense_activation_kwargs)
                               ]}
                sequential_layer = {"type": "Sequential", "kwargs": layers}
                dense_hparams.append(sequential_layer)

            elif (i == ndense - 1 and
                  self._hparams.final_dense_activation is not None):
                layers = {
                    "layers": [dense_hparams_i,
                               _activation_hparams(
                                   self._hparams.final_dense_activation,
                                   self._hparams.final_dense_activation_kwargs)
                               ]}
                sequential_layer = {"type": "Sequential", "kwargs": layers}
                dense_hparams.append(sequential_layer)
            else:
                dense_hparams.append(dense_hparams_i)

        def _dropout_hparams():
            return {"type": "Dropout",
                    "kwargs": {"p": self._hparams.dropout_rate}}

        dropout_dense = _to_list(self._hparams.dropout_dense)
        ndense = self._hparams.num_dense_layers
        if ndense > 0:  # Add flatten layers before dense layers
            layer_hparams.append({"type": "Flatten"})
        for dense_i in range(ndense):
            if dense_i in dropout_dense:
                layer_hparams.append(_dropout_hparams())
            layer_hparams.append(dense_hparams[dense_i])
        if ndense in dropout_dense:
            layer_hparams.append(_dropout_hparams())

        return layer_hparams

    def _build_non_dense_layer_hparams(self, in_channels):
        pool_hparams = self._build_pool_hparams()
        conv_pool_hparams = self._build_conv1d_hparams(in_channels,
                                                       pool_hparams)

        def _dropout_hparams():
            return {"type": "Dropout",
                    "kwargs": {"p": self._hparams.dropout_rate}}

        dropout_conv = _to_list(self._hparams.dropout_conv)

        layers_hparams = []
        nconv = self._hparams.num_conv_layers
        for conv_i in range(nconv):
            if conv_i in dropout_conv:
                layers_hparams.append(_dropout_hparams())
            if isinstance(conv_pool_hparams[conv_i], (list, tuple)):
                layers_hparams += conv_pool_hparams[conv_i]
            else:
                layers_hparams.append(conv_pool_hparams[conv_i])
        if nconv in dropout_conv:
            layers_hparams.append(_dropout_hparams())

        return layers_hparams

    def forward(self,  # type: ignore
                input: torch.Tensor,
                sequence_length: Optional[Union[torch.LongTensor,
                                                List[int]]] = None,
                dtype: Optional[torch.dtype] = None,
                data_format: Optional[str] = None) -> torch.Tensor:
        r"""Feeds forward inputs through the network layers and returns outputs.

        Args:
            input: The inputs to the network, which is a 3D tensor.
            sequence_length (optional): An :tensor:`LongTensor` of shape
                ``[batch_size]`` or a python array containing the length of
                each element in :attr:`inputs`. If given, time steps beyond
                the length will first be masked out before feeding to the
                layers.
            dtype (optional): Type of the inputs. If not provided,
                infers from inputs automatically.
            data_format (optional): Data type of the input tensor. If
                ``channels_last``, the last dimension will be treated as channel
                dimension so the size of the :attr:`input` should be
                `[batch_size, X, channel]`. If ``channels_first``, first
                dimension will be treated as channel dimension so the size
                should be `[batch_size, channel, X]`. Defaults to None.
                If None, the value will be picked from hyperparameters.
        Returns:
            The output of the final layer.
        """
        if input.dim() != 3:
            raise ValueError("'input' should be a 3D tensor.")

        if data_format is None:
            data_format = self.hparams["data_format"]

        if data_format == "channels_first":
            # masking requires channels in last dimension
            input = input.permute(0, 2, 1)

            if sequence_length is not None:
                input = mask_sequences(input, sequence_length,
                                       dtype=dtype, time_major=False)

            # network is constructed for channel first tensors
            input = input.permute(0, 2, 1)

            output = super().forward(input)

        elif data_format == "channels_last":
            if sequence_length is not None:
                input = mask_sequences(input, sequence_length,
                                       dtype=dtype, time_major=False)

            input = input.permute(0, 2, 1)

            output = super().forward(input)

            # transpose only when tensors are 3D
            if output.dim() == 3:
                output = output.permute(0, 2, 1)
        else:
            raise ValueError("Invalid 'data_format'")

        return output

    def _infer_dense_layer_input_size(self, input: torch.Tensor) -> torch.Size:
        # feed forward the input on the conv part of the network to infer
        # input shape for dense layers
        with torch.no_grad():
            output = super().forward(input)
        return output.view(output.size()[0], -1).size()

    @property
    def output_size(self) -> int:
        r"""The feature size of :meth:`forward` output.
        """
        if self.hparams.num_dense_layers <= 0:
            out_channels = self._hparams.out_channels
            if not isinstance(out_channels, (list, tuple)):
                out_channels = [out_channels]
            nconv = self._hparams.num_conv_layers
            if nconv == 1:
                kernel_size = _to_list(self._hparams.kernel_size)
                if not isinstance(kernel_size[0], (list, tuple)):
                    kernel_size = [kernel_size]
            elif nconv > 1:
                kernel_size = _to_list(self._hparams.kernel_size,
                                       'kernel_size', nconv)
                kernel_size = [_to_list(ks) for ks in kernel_size]
            return out_channels[-1] * len(kernel_size[-1])
        else:
            out_features = self._hparams.out_features
            if isinstance(out_features, (list, tuple)):
                return out_features[-1]
            else:
                return out_features



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for pre-trained BERT tokenizer.
"""

import unittest

import os
import pickle
import tempfile

from texar.torch.data.tokenizers.bert_tokenizer import \
    BERTTokenizer
from texar.torch.utils.test import pretrained_test


class BERTTokenizerTest(unittest.TestCase):

    def setUp(self):
        vocab_tokens = [
            "[UNK]", "[CLS]", "[SEP]", "want", "##want", "##ed", "wa", "un",
            "runn",
            "##ing", ",", "low", "lowest",
        ]

        self.tmp_dir = tempfile.TemporaryDirectory()
        self.vocab_file = os.path.join(self.tmp_dir.name, 'vocab.txt')
        with open(self.vocab_file, "w", encoding='utf-8') as vocab_writer:
            vocab_writer.write("".join([x + "\n" for x in vocab_tokens]))

    def tearDown(self):
        self.tmp_dir.cleanup()

    @pretrained_test
    def test_model_loading(self):
        for pretrained_model_name in BERTTokenizer.available_checkpoints():
            tokenizer = BERTTokenizer(
                pretrained_model_name=pretrained_model_name)
            _ = tokenizer.map_text_to_token(u"UNwant\u00E9d,running")

    def test_tokenize(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)

        tokens = tokenizer.map_text_to_token(u"UNwant\u00E9d,running")
        self.assertListEqual(tokens,
                             ["un", "##want", "##ed", ",", "runn", "##ing"])

        ids = tokenizer.map_token_to_id(tokens)
        self.assertListEqual(ids, [7, 4, 5, 10, 8, 9])

    def test_pickle(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)
        self.assertIsNotNone(tokenizer)

        text = u"Munich and Berlin are nice cities"
        subwords = tokenizer.map_text_to_token(text)

        with tempfile.TemporaryDirectory() as tmpdirname:
            filename = os.path.join(tmpdirname, u"tokenizer.bin")
            with open(filename, "wb") as f:
                pickle.dump(tokenizer, f)
            with open(filename, "rb") as f:
                tokenizer_new = pickle.load(f)

        subwords_loaded = tokenizer_new.map_text_to_token(text)

        self.assertListEqual(subwords, subwords_loaded)

    def test_save_load(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)

        before_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")

        with tempfile.TemporaryDirectory() as tmpdirname:
            tokenizer.save(tmpdirname)
            tokenizer = tokenizer.load(tmpdirname)

        after_tokens = tokenizer.map_text_to_id(
            u"He is very happy, UNwant\u00E9d,running")
        self.assertListEqual(before_tokens, after_tokens)

    def test_pretrained_model_list(self):
        model_list_1 = list(BERTTokenizer._MODEL2URL.keys())
        model_list_2 = list(BERTTokenizer._MAX_INPUT_SIZE.keys())

        self.assertListEqual(model_list_1, model_list_2)

    def test_encode_decode(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)

        input_text = u"UNwant\u00E9d,running"
        output_text = u"unwanted, running"

        tokens = tokenizer.map_text_to_token(input_text)
        ids = tokenizer.map_token_to_id(tokens)
        ids_2 = tokenizer.map_text_to_id(input_text)
        self.assertListEqual(ids, ids_2)

        tokens_2 = tokenizer.map_id_to_token(ids)
        text_2 = tokenizer.map_id_to_text(ids)

        self.assertEqual(text_2, output_text)

        self.assertNotEqual(len(tokens_2), 0)
        self.assertIsInstance(text_2, str)

    def test_add_tokens(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)

        vocab_size = tokenizer.vocab_size
        all_size = len(tokenizer)

        self.assertNotEqual(vocab_size, 0)
        self.assertEqual(vocab_size, all_size)

        new_toks = ["aaaaabbbbbb", "cccccccccdddddddd"]
        added_toks = tokenizer.add_tokens(new_toks)
        vocab_size_2 = tokenizer.vocab_size
        all_size_2 = len(tokenizer)

        self.assertNotEqual(vocab_size_2, 0)
        self.assertEqual(vocab_size, vocab_size_2)
        self.assertEqual(added_toks, len(new_toks))
        self.assertEqual(all_size_2, all_size + len(new_toks))

        tokens = tokenizer.map_text_to_id("aaaaabbbbbb low cccccccccdddddddd l")
        self.assertGreaterEqual(len(tokens), 4)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)

        new_toks_2 = {'eos_token': ">>>>|||<||<<|<<",
                      'pad_token': "<<<<<|||>|>>>>|>"}
        added_toks_2 = tokenizer.add_special_tokens(new_toks_2)
        vocab_size_3 = tokenizer.vocab_size
        all_size_3 = len(tokenizer)

        self.assertNotEqual(vocab_size_3, 0)
        self.assertEqual(vocab_size, vocab_size_3)
        self.assertEqual(added_toks_2, len(new_toks_2))
        self.assertEqual(all_size_3, all_size_2 + len(new_toks_2))

        tokens = tokenizer.map_text_to_id(
            ">>>>|||<||<<|<< aaaaabbbbbb low cccccccccdddddddd "
            "<<<<<|||>|>>>>|> l")

        self.assertGreaterEqual(len(tokens), 6)
        self.assertGreater(tokens[0], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[0], tokens[1])
        self.assertGreater(tokens[-2], tokenizer.vocab_size - 1)
        self.assertGreater(tokens[-2], tokens[-3])
        self.assertEqual(tokens[0],
                         tokenizer.map_token_to_id(tokenizer.eos_token))
        self.assertEqual(tokens[-2],
                         tokenizer.map_token_to_id(tokenizer.pad_token))

    def test_encode_text(self):
        tokenizer = BERTTokenizer.load(self.vocab_file)

        text_1 = u"He is very happy"
        text_2 = u"unwanted, running"

        text_1_ids = tokenizer.map_text_to_id(text_1)
        text_2_ids = tokenizer.map_text_to_id(text_2)

        cls_token_id = tokenizer.map_token_to_id(tokenizer.cls_token)
        sep_token_id = tokenizer.map_token_to_id(tokenizer.sep_token)

        input_ids, segment_ids, input_mask = \
            tokenizer.encode_text(text_1, None, 4)

        self.assertListEqual(input_ids,
                             [cls_token_id] + text_1_ids[:2] + [sep_token_id])
        self.assertListEqual(segment_ids, [0, 0, 0, 0])
        self.assertListEqual(input_mask, [1, 1, 1, 1])

        input_ids, segment_ids, input_mask = \
            tokenizer.encode_text(text_1, text_2, 7)

        self.assertListEqual(input_ids, [cls_token_id] + text_1_ids[:2] +
                             [sep_token_id] + text_2_ids[:2] + [sep_token_id])
        self.assertListEqual(segment_ids, [0, 0, 0, 0, 1, 1, 1])
        self.assertListEqual(input_mask, [1, 1, 1, 1, 1, 1, 1])


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions for the Executor module.
"""

import functools
import time
from collections import Counter, OrderedDict
from typing import (
    Any, Callable, Counter as CounterType, Dict, Iterator, List, NamedTuple,
    Optional, Tuple, Type, TypeVar, Union, Mapping, Sequence)

from mypy_extensions import TypedDict
import torch
from torch import nn

from texar.torch.data.data.dataset_utils import Batch
from texar.torch.run.metric import Metric
from texar.torch.utils.types import MaybeSeq
from texar.torch.utils.utils import get_instance

__all__ = [
    "OptionalList",
    "OptionalDict",
    "Instance",
    "to_list",
    "to_metric_dict",
    "to_instance",
    "SavedTrainingState",
    "TrainingStatus",
    "CheckpointMetaInfo",
    "ProgressTracker",
    "ExecutorTerminateSignal",
    "MetricList",
    "update_metrics",
    "color",
    "repr_module",
]

T = TypeVar('T')
OptionalList = Optional[MaybeSeq[T]]
OptionalDict = Optional[Union[T, Sequence[Union[T, Tuple[str, T]]],
                              Mapping[str, T]]]
Instance = Union[T, Dict[str, Any]]


def to_list(xs: OptionalList[T]) -> List[T]:
    if isinstance(xs, Sequence):
        return list(xs)
    if xs is None:
        return []
    return [xs]


def _to_dict(ds: OptionalDict[T],
             unambiguous_name_fn: Callable[[str, T, int], str],
             default_name_fn: Callable[[int, T], str]) -> 'OrderedDict[str, T]':
    if ds is None:
        return OrderedDict()
    if isinstance(ds, Mapping):
        return OrderedDict(ds)
    if isinstance(ds, Sequence):
        xs = ds
    else:
        xs = [ds]
    ret_dict: 'OrderedDict[str, T]' = OrderedDict()
    counter: CounterType[str] = Counter()
    for idx, item in enumerate(xs):
        if isinstance(item, tuple):
            name, item = item
        else:
            name = default_name_fn(idx, item)
        if name not in counter:
            ret_dict[name] = item
        else:
            cnt = counter[name]
            if cnt == 1:
                prev_item = ret_dict[name]
                ret_dict[unambiguous_name_fn(name, prev_item, 1)] = prev_item
                del ret_dict[name]
            ret_dict[unambiguous_name_fn(name, item, cnt + 1)] = item
        counter.update([name])
    return ret_dict


def to_dict(xs: OptionalDict[T],
            default_name: Optional[str] = None) -> Dict[str, T]:
    def unambiguous_name_fn(name: str, _item: T, cnt: int) -> str:
        return f"{name}.{cnt}"

    def default_name_fn(idx: int, _item: T) -> str:
        if default_name is not None:
            return default_name
        return str(idx)

    return _to_dict(xs, unambiguous_name_fn, default_name_fn)


def to_metric_dict(metrics: OptionalDict[Metric]) -> 'OrderedDict[str, Metric]':
    def unambiguous_name_fn(name: str, metric: Metric, _cnt: int) -> str:
        new_name = f"{name}_{metric.pred_name}"
        if metric.label_name is not None:
            new_name = f"{new_name}_{metric.label_name}"
        return new_name

    def default_name_fn(_idx: int, metric: Metric) -> str:
        return metric.metric_name

    if isinstance(metrics, dict) and not isinstance(metrics, OrderedDict):
        raise ValueError("Metrics dictionary must be of type OrderedDict")
    metric_dict = _to_dict(metrics, unambiguous_name_fn, default_name_fn)

    for name, metric in metric_dict.items():
        if not name.isidentifier():
            raise ValueError(f"Name \"{name}\" for metric {metric} is not a "
                             f"valid identifier name")
        if not isinstance(metric, Metric):
            raise ValueError(f"All metrics must be of class Metric, but found "
                             f"{type(metric)}")
    return metric_dict


def to_instance(typ: Type[T], instance: Instance[T], modules: List[str],
                extra_kwargs: Optional[Dict[str, Any]] = None) -> Optional[T]:
    if instance is None:
        return None
    if isinstance(instance, dict):
        kwargs = {**instance.get('kwargs', {}), **(extra_kwargs or {})}
        instance = get_instance(instance['type'], kwargs, modules)
    if not isinstance(instance, typ):
        raise ValueError(f"The instance {instance} is not of type {typ}")
    return instance


# TODO: Also save training progress?
class SavedTrainingState(NamedTuple):
    r"""The entire training state to save to or load from checkpoints."""
    model: Dict[str, torch.Tensor]
    optimizer: Dict[str, torch.Tensor]
    scheduler: Optional[Dict[str, Any]]
    system_rng: Any
    numpy_rng: Any
    torch_rng: Any


class TrainingStatus(TypedDict):
    epoch: int
    iteration: int
    split: str
    metric: 'OrderedDict[str, Metric]'
    eval_metric: 'OrderedDict[str, Metric]'


class CheckpointMetaInfo(TypedDict):
    status: TrainingStatus
    timestamp: float


class ProgressTracker:
    start_time: float
    size: Optional[int]
    n_examples: int
    accumulated_time: float
    paused: bool

    _tracker_stack: List['ProgressTracker'] = []

    def __init__(self, size: Optional[int] = None):
        self.size = size
        self.started = False

    def set_size(self, size: Optional[int]):
        self.size = size

    def start(self):
        if self.started:
            return
        self.started = True
        if len(self._tracker_stack) > 0:
            self._tracker_stack[-1].pause()
        self._tracker_stack.append(self)
        self.reset()
        self.paused = False

    def stop(self):
        if not self.started:
            return
        self.started = False
        obj = self._tracker_stack.pop(-1)
        assert obj is self
        if len(self._tracker_stack) > 0:
            self._tracker_stack[-1].resume()

    def reset(self):
        self.n_examples = 0
        self.start_time = time.time()
        self.accumulated_time = 0.0

    def pause(self):
        if self.paused:
            return
        self.paused = True
        self.accumulated_time += time.time() - self.start_time

    def resume(self):
        if not self.paused:
            return
        self.paused = False
        self.start_time = time.time()

    def add(self, n_examples: int):
        self.n_examples += n_examples

    def progress(self) -> Optional[float]:
        if self.size is None:
            return None
        return self.n_examples / self.size * 100

    def time_elapsed(self) -> float:
        return self.accumulated_time + (time.time() - self.start_time)

    def speed(self) -> str:
        speed = self.n_examples / self.time_elapsed()
        if speed > 1.0:
            return f"{speed:.2f}ex/s"
        try:
            return f"{1.0 / speed:.2f}s/ex"
        except ZeroDivisionError:
            return f"0.00ex/s"


class ExecutorTerminateSignal(Exception):
    pass


@functools.total_ordering
class MetricList:
    r"""A class representing list of metrics along with their values at a
    certain point. Used for metric comparisons.

    Args:
        metrics: The dictionary of metric instances.
        values (optional): The dictionary of metric values. If `None` (default),
            the current values of the provided metrics are used.
    """

    # TODO: Ignore non-streaming metrics here? Or in

    def __init__(self, metrics: 'OrderedDict[str, Metric]',
                 values: Optional[Dict[str, Any]] = None):
        self.metrics = metrics
        if values is None:
            self.values = {name: metric.value()
                           for name, metric in metrics.items()}
        else:
            self.values = values

    def _compare_metrics(self, other: Any):
        if not isinstance(other, MetricList):
            raise ValueError(
                "Cannot compare to an object not of type MetricList")
        for (name, metric), (other_name, other_metric) in zip(
                self.metrics.items(), other.metrics.items()):
            if name != other_name or type(metric) is not type(other_metric):
                raise ValueError("Cannot compare two metric lists with "
                                 "different base metrics")

    def __eq__(self, other: Any) -> bool:
        self._compare_metrics(other)
        return all(self.values[name] == other.values[name]
                   for name in self.metrics)

    def __gt__(self, other: 'MetricList') -> bool:
        r"""Compare this metric list to another, and return whether the current
        list is better.
        """
        self._compare_metrics(other)
        for name, metric in self.metrics.items():
            cmp = metric.better(self.values[name], other.values[name])
            if cmp is not None:
                return cmp
        return False


def update_metrics(return_dict: Dict[str, Any], batch: Batch,
                   metrics: 'OrderedDict[str, Metric]') -> None:
    for metric_name, metric in metrics.items():
        if metric.pred_name is not None:
            try:
                pred_val = return_dict[metric.pred_name]
            except KeyError:
                raise ValueError(
                    f"Return dictionary from model does not contain "
                    f"'{metric.pred_name}' entry, which was required for "
                    f"metric '{metric_name}'")
            if isinstance(pred_val, torch.Tensor):
                pred_val = pred_val.tolist()
            pred_val = to_list(pred_val)
        else:
            pred_val = None
        if metric.label_name is not None:
            try:
                label_val = batch[metric.label_name]
            except KeyError:
                raise ValueError(
                    f"Data batch does not contain '{metric.label_name}' "
                    f"entry, which was required for metric '{metric_name}'")
            if isinstance(label_val, torch.Tensor):
                label_val = label_val.tolist()
            label_val = to_list(label_val)
        else:
            label_val = None
        metric.add(pred_val, label_val)


CLEAR_LINE = '\033[2K\r'
RESET_CODE = '\033[0m'
COLOR_CODE = {
    'red': '\033[31m',
    'green': '\033[32m',
    'yellow': '\033[33m',
    'blue': '\033[94m',
    'magenta': '\033[35m',
    'cyan': '\033[36m',
    'gray': '\033[37m',
    'grey': '\033[37m'
}


def color(s: str, col: str):
    return COLOR_CODE[col.lower()] + s + RESET_CODE


def _add_indent(s_, n_spaces):
    s = s_.split('\n')
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    first = s.pop(0)
    s = [(n_spaces * ' ') + line for line in s]
    s = '\n'.join(s)
    s = first + '\n' + s
    return s


def _convert_id(keys: List[str]) -> Iterator[str]:
    start = end = None
    for key in keys:
        if key.isnumeric() and end == int(key) - 1:
            end = int(key)
        else:
            if start is not None:
                if start == end:
                    yield f"id {start}"
                else:
                    yield f"ids {start}-{end}"
            if key.isnumeric():
                start = end = int(key)
            else:
                start = end = None
                yield key
    if start is not None:
        if start == end:
            yield f"id {start}"
        else:
            yield f"ids {start}-{end}"


def repr_module(module: nn.Module) -> str:
    r"""Create a compressed representation by combining identical modules in
    `nn.ModuleList`s and `nn.ParameterList`s.
    """

    # We treat the extra repr like the sub-module, one item per line
    extra_lines: List[str] = []
    extra_repr = module.extra_repr()
    # empty string will be split into list ['']
    if extra_repr:
        extra_lines = extra_repr.split('\n')
    child_lines = []
    prev_mod_str = None
    keys: List[str] = []
    for key, submodule in module.named_children():
        mod_str = repr_module(submodule)
        mod_str = _add_indent(mod_str, 2)
        if prev_mod_str is None or prev_mod_str != mod_str:
            if prev_mod_str is not None:
                for name in _convert_id(keys):
                    child_lines.append(f"({name}): {prev_mod_str}")
            prev_mod_str = mod_str
            keys = [key]
        else:
            keys.append(key)
    if len(keys) > 0:
        for name in _convert_id(keys):
            child_lines.append(f"({name}): {prev_mod_str}")
    lines = extra_lines + child_lines

    main_str = module.__class__.__name__ + '('
    if lines:
        # simple one-liner info, which most builtin Modules will use
        if len(extra_lines) == 1 and not child_lines:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(lines) + '\n'

    main_str += ')'
    return main_str



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for XLNet decoder.
"""
import unittest

import torch

from texar.torch.modules.decoders.xlnet_decoder import *
from texar.torch.utils.test import pretrained_test


class XLNetDecoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.XLNetDecoder`
    """

    def setUp(self) -> None:
        self.batch_size = 2
        self.max_length = 3
        self.start_tokens = torch.zeros(
            self.batch_size, self.max_length, dtype=torch.long)

    @pretrained_test
    def test_hparams(self):
        r"""Tests the priority of the decoder arch parameters.
        """
        # case 1: set "pretrained_mode_name" by constructor argument
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
        }
        decoder = XLNetDecoder(pretrained_model_name="xlnet-base-cased",
                               hparams=hparams)
        self.assertEqual(decoder.hparams.num_layers, 12)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

        # case 2: set "pretrained_mode_name" by hparams
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
            "num_layers": 6,
        }
        decoder = XLNetDecoder(hparams=hparams)
        self.assertEqual(decoder.hparams.num_layers, 24)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

        # case 3: set to None in both hparams and constructor argument
        hparams = {
            "pretrained_model_name": None,
            "num_layers": 6,
        }
        decoder = XLNetDecoder(hparams=hparams)
        self.assertEqual(decoder.hparams.num_layers, 6)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

        # case 4: using default hparams
        decoder = XLNetDecoder()
        self.assertEqual(decoder.hparams.num_layers, 12)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

    @pretrained_test
    def test_trainable_variables(self):
        r"""Tests the functionality of automatically collecting trainable
        variables.
        """
        # case 1
        decoder = XLNetDecoder()
        self.assertEqual(len(decoder.trainable_variables), 182 + 1)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

        # case 2
        hparams = {
            "pretrained_model_name": "xlnet-large-cased",
        }
        decoder = XLNetDecoder(hparams=hparams)
        self.assertEqual(len(decoder.trainable_variables), 362 + 1)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

        # case 3
        hparams = {
            "pretrained_model_name": None,
            "num_layers": 6
        }
        decoder = XLNetDecoder(hparams=hparams)
        self.assertEqual(len(decoder.trainable_variables), 92 + 1)

        _, _ = decoder(start_tokens=self.start_tokens,
                       end_token=1, max_decoding_length=self.max_length)

    @pretrained_test
    def test_decode_infer_sample(self):
        r"""Tests train_greedy."""
        hparams = {
            "pretrained_model_name": None,
        }
        decoder = XLNetDecoder(hparams=hparams)
        decoder.train()

        inputs = torch.randint(32000, (self.batch_size, self.max_length))
        outputs, _ = decoder(
            inputs, max_decoding_length=self.max_length, end_token=2)

        self.assertIsInstance(outputs, XLNetDecoderOutput)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for connectors that transform inputs into specified output shape.
"""

from abc import ABC
from typing import Generic, Optional, TypeVar
from texar.torch.hyperparams import HParams
from texar.torch.module_base import ModuleBase

__all__ = [
    "ConnectorBase"
]

OutputSize = TypeVar('OutputSize')
HParamsType = Optional[HParams]


class ConnectorBase(ModuleBase, Generic[OutputSize], ABC):
    r"""Base class inherited by all connector classes. A connector is to
    transform inputs into outputs with any specified structure and shape.
    For example, transforming the final state of an encoder to the initial
    state of a decoder, and performing stochastic sampling in between as
    in Variational Autoencoders (VAEs).

    Args:
        output_size: Size of output **excluding** the batch dimension. For
            example, set ``output_size`` to ``dim`` to generate output of
            shape ``[batch_size, dim]``.
            Can be an `int`, a tuple of `int`, a torch.Size, or a tuple of
            torch.Sizes.
            For example, to transform inputs to have decoder state size, set
            :python:`output_size=decoder.state_size`.
        hparams (dict, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure and
            default values.
    """

    def __init__(self, output_size: OutputSize, hparams: HParamsType = None):
        super().__init__(hparams=hparams)
        self._output_size = output_size

    @staticmethod
    def default_hparams():
        r"""Returns a dictionary of hyperparameters with default values.
        """
        return {
            "name": "connector"
        }

    @property
    def output_size(self):
        return self._output_size



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various optimization related utilities.
"""

import functools
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from mypy_extensions import TypedDict

import torch
import torch.nn as nn
from torch.nn.utils import clip_grad_norm_
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

from texar.torch.hyperparams import HParams
from texar.torch.utils import utils
from texar.torch.utils.types import MaybeList

__all__ = [
    "default_optimization_hparams",
    "get_optimizer",
    "get_scheduler",
    "get_grad_clip_fn",
    "get_train_op",
    "BertAdam"
]


def default_optimization_hparams() -> Dict[str, Any]:
    r"""Returns a `dict` of default hyperparameters of training op
    and their default values

    .. code-block:: python

        {
            "optimizer": {
                "type": "Adam",
                "kwargs": {
                    "lr": 0.001
                }
            },
            "learning_rate_decay": {
                "type": "",
                "kwargs": {}
            },
            "gradient_clip": {
                "type": "",
                "kwargs": {}
            },
            "gradient_noise_scale": None,
            "name": None
        }

    Here:

    `"optimizer"`: dict
        Hyperparameters of a
        :torch_docs:`torch.optim.Optimizer <optim.html#torch.optim.Optimizer>`.

        - `"type"` specifies the optimizer class. This can be

          - The string name or full module path of an optimizer class.
            If the class name is provided, the class must be in module
            :torch_docs:`torch.optim <optim.html>` or :mod:`texar.torch.custom`,
            :mod:`texar.torch.core.optimization`
          - An optimizer class.
          - An instance of an optimizer class.

          For example

          .. code-block:: python

              "type": "Adam"                    # class name
              "type": "my_module.MyOptimizer"   # module path
              "type": texar.torch.custom.BertAdam     # class
              "type": my_module.MyOptimizer     # class

        - `"kwargs"` is a `dict` specifying keyword arguments for creating
          the optimizer class instance, with :python:`opt_class(**kwargs)`.
          Ignored if `"type"` is a class instance.

    `"learning_rate_decay"`: dict
        Hyperparameters of learning rate decay function. The learning rate
        starts decay from :attr:`"start_decay_step"` and keeps unchanged after
        :attr:`"end_decay_step"` or reaching :attr:`"min_learning_rate"`.

        The decay function is specified in `"type"` and `"kwargs"`.

        - `"type"` can be a decay function or its name or module path. If
          function name is provided, it must be from module
          :torch_docs:`torch.optim <optim.html>` or :mod:`texar.torch.custom`,
          :mod:`texar.torch.core.optimization`.

        - `"kwargs"` is a `dict` of keyword arguments for the function
          excluding arguments named `"global_step"` and `"learning_rate"`.

        The function is called with
        :python:`lr = decay_fn(learning_rate=lr, global_step=offset_step,
        **kwargs)`, where `offset_step` is the global step offset as above.

    `"gradient_clip"`: dict
        Hyperparameters of gradient clipping. The gradient clipping function
        takes a list of `(gradients, variables)` tuples and returns a list
        of `(clipped_gradients, variables)` tuples. Typical examples include
        :torch_nn:`utils.clip_grad_norm_` and
        :torch_nn:`utils.clip_grad_value_`.

        "type" specifies the gradient clip function, and can be a function,
        or its name or module path. If function name is provided, the
        function must be from module :mod:`torch.nn.utils`,
        :mod:`texar.torch.custom`, or :mod:`texar.torch.core.optimization`.

        `"kwargs"` specifies keyword arguments to the function, except arguments
        named `"parameters"`.

    `"gradient_noise_scale"`: float, optional
        Adds 0-mean normal noise scaled by this value to gradient.
    """
    return {
        "optimizer": {
            "type": "Adam",
            "kwargs": {
                "lr": 0.001
            }
        },
        "learning_rate_decay": {
            "type": "",
            "kwargs": {}
        },
        "gradient_clip": {
            "type": "",
            "kwargs": {}
        },
        "gradient_noise_scale": None,
        # TODO(zhiting): allow module-level control of gradient_multipliers
        "name": None
    }


def get_optimizer(
        params: Iterable[Union[torch.Tensor, Dict[str, Any]]],
        hparams: Optional[Union[HParams, Dict[str, Any]]] = None) -> \
        Optimizer:
    r"""Creates a optimizer instance.

    Args:
        params: an iterable of :class:`torch.Tensor` or
            :class:`dict`. Specifies what Tensors should be optimized.
        hparams (dict or HParams, optional): hyperparameters. Missing
            hyperparameters are set to default values automatically. See
            :func:`~texar.torch.core.default_optimization_hparams` for
            all hyperparameters and default values.

    :return:
        The :torch_docs:`torch.optim.Optimizer
        <optim.html#torch.optim.Optimizer>` instance specified in
        :attr:`hparams`.
    """
    if hparams is None or isinstance(hparams, dict):
        hparams = HParams(hparams, default_optimization_hparams())

    hparams_opt = hparams["optimizer"]

    optimizer_type = hparams_opt["type"]
    if isinstance(optimizer_type, Optimizer):
        optimizer_class = optimizer_type
    else:
        optimizer_modules = ['torch.optim',
                             'texar.torch.custom']
        try:
            optimizer_class = utils.check_or_get_class(  # type: ignore
                optimizer_type, optimizer_modules, Optimizer)
        except TypeError:
            raise ValueError(
                "Unrecognized optimizer. Must be string name of the "
                "optimizer class, or the class which is a subclass of "
                "torch.optim.Optimizer, or an instance of the subclass of "
                "Optimizer.")

    optimizer_kwargs = hparams_opt["kwargs"].todict()
    optimizer_kwargs.update({"params": params})
    optimizer = optimizer_class(**optimizer_kwargs)  # type: ignore

    return optimizer


def get_scheduler(optimizer: Optimizer,
                  hparams: Optional[Union[HParams, Dict[str, Any]]] = None) -> \
        Optional[_LRScheduler]:
    r"""Creates a scheduler instance.

    Args:
        optimizer: A :torch_docs:`torch.optim.Optimizer
            <optim.html#torch.optim.Optimizer>` instance.
        hparams (dict or HParams, optional): hyperparameters. Missing
            hyperparameters are set to default values automatically. See
            :func:`~texar.torch.core.default_optimization_hparams` for
            all hyperparameters and default values.

    :return:
        A :torch_docs:`torch.optim.lr_scheduler._LRScheduler
        <optim.html#how-to-adjust-learning-rate>` instance.
    """
    if hparams is None or isinstance(hparams, dict):
        hparams = HParams(hparams, default_optimization_hparams())

    hparams_scheduler = hparams["learning_rate_decay"]

    scheduler_type = hparams_scheduler["type"]
    if scheduler_type == "" or scheduler_type is None:
        scheduler = None
    else:
        if isinstance(scheduler_type, _LRScheduler):
            scheduler_class = scheduler_type
        else:
            scheduler_modules = ['torch.optim.lr_scheduler',
                                 'texar.torch.custom']
            try:
                scheduler_class = utils.check_or_get_class(  # type: ignore
                    scheduler_type, scheduler_modules, _LRScheduler)
            except TypeError:
                raise ValueError(
                    "Unrecognized lr_scheduler. Must be string name of the "
                    "lr_scheduler class, or the class which is a subclass of "
                    "torch.optim._LRScheduler.")

        scheduler_kwargs = hparams_scheduler["kwargs"].todict()
        scheduler_kwargs.update({"optimizer": optimizer})
        scheduler = scheduler_class(**scheduler_kwargs)  # type: ignore

    return scheduler


def get_grad_clip_fn(hparams: Optional[Union[HParams,
                                             Dict[str, Any]]] = None) -> \
        Optional[Callable[[torch.Tensor], Optional[torch.Tensor]]]:
    r"""Create a gradient clipping function.

    Args:
        hparams (dict or HParams, optional): hyperparameters. Missing
            hyperparameters are set to default values automatically. See
            :func:`~texar.torch.core.default_optimization_hparams` for
            all hyperparameters and default values.

    Returns:
        A gradient clipping function.
    """
    if hparams is None or isinstance(hparams, dict):
        hparams = HParams(hparams, default_optimization_hparams())

    hparams_grad_clip = hparams["gradient_clip"]

    grad_clip_type = hparams_grad_clip["type"]
    if grad_clip_type == "" or grad_clip_type is None:
        grad_clip_fn = None
    else:
        grad_clip_modules = ['torch.nn.utils',
                             'texar.torch.custom']
        grad_clip_fn = utils.get_function(grad_clip_type, grad_clip_modules)
        grad_clip_fn_kwargs = hparams_grad_clip["kwargs"].todict()
        grad_clip_fn = functools.partial(grad_clip_fn, **grad_clip_fn_kwargs)

    return grad_clip_fn


def get_train_op(params: Optional[Iterable[Union[torch.Tensor,
                                                 Dict[str, Any]]]] = None,
                 optimizer: Optional[Optimizer] = None,
                 scheduler: Optional[_LRScheduler] = None,
                 hparams: Optional[Union[HParams, Dict[str, Any]]] = None) -> \
        Callable[[], None]:
    r"""Creates a training op.

    Args:
        params: an iterable of :class:`torch.Tensor` or
            :class:`dict`. Specifies what Tensors should be optimized.
        optimizer: A :torch_docs:`torch.optim.Optimizer
            <optim.html#torch.optim.Optimizer>` instance.
        scheduler: A :torch_docs:`torch.optim.lr_scheduler._LRScheduler
            <optim.html#how-to-adjust-learning-rate>` instance.
        hparams (dict or HParams, optional): hyperparameters. Missing
            hyperparameters are set to default values automatically. See
            :func:`~texar.torch.core.default_optimization_hparams` for
            all hyperparameters and default values.

    Returns:
        The callable used for variable optimization.
    """
    hparams = HParams(hparams, default_optimization_hparams())

    if params is None and optimizer is None and scheduler is None:
        raise ValueError("'params', 'optimizer' and 'scheduler' must not be "
                         "None simultaneously.")

    if scheduler is None:
        if optimizer is None and params is not None:
            optimizer = get_optimizer(params, hparams)
        if optimizer is not None:
            scheduler = get_scheduler(optimizer, hparams)
    else:
        optimizer = scheduler.optimizer  # type: ignore

    grad_clip_fn = get_grad_clip_fn(hparams)

    # TODO: Support per-parameter options in the future.
    params_list: List[nn.Parameter] = []
    for param_group in optimizer.param_groups:  # type: ignore
        params = param_group["params"]
        if isinstance(params, torch.Tensor):
            params_list.append(params)
        elif isinstance(params, list):
            params_list += params

    def _train_op():
        if grad_clip_fn is not None:
            grad_clip_fn(parameters=params_list)
        optimizer.step()
        # TODO: Ideally, scheduler should be used in the epoch level.
        if scheduler is not None:
            scheduler.step()
        optimizer.zero_grad()

    return _train_op


class BertAdamParamDict(TypedDict):
    r"""The :attr:`param_groups` dictionary used in PyTorch optimizers."""
    params: List[nn.Parameter]
    lr: float
    betas: Tuple[float, float]
    eps: float
    weight_decay: float
    max_grad_norm: float


class BertAdamStateDict(TypedDict):
    r"""The :attr:`state` dictionary used in :class:`BertAdam` optimizer."""
    next_m: torch.Tensor
    next_v: torch.Tensor


OptimParamType = Union[
    MaybeList[Iterable[nn.Parameter]],  # model.parameters()
    MaybeList[Dict[str, Any]],  # {"params": ..., "other_kwargs": ...}
]


class BertAdam(Optimizer):
    r"""Implements BERT version of Adam algorithm with weight decay fix.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        max_grad_norm: Maximum norm for the gradients (-1 means no clipping).
            Default: 1.0
    """

    param_groups: List[BertAdamParamDict]
    state: Dict[nn.Parameter, BertAdamStateDict]

    def __init__(self, params: OptimParamType,
                 lr: float = 0.001, betas: Tuple[float, float] = (0.9, 0.999),
                 eps: float = 1e-08, weight_decay: float = 0,
                 max_grad_norm: float = 1.0):

        if lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if eps < 0.0:
            raise ValueError(f"Invalid epsilon value: {eps}")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 0: {betas[0]}")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter at index 1: {betas[1]}")

        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, max_grad_norm=max_grad_norm)
        super().__init__(params, defaults)  # type: ignore

    def step(self, closure: Optional[Callable[[], float]] = None):
        r"""Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError(
                        "Adam does not support sparse gradients, please "
                        "consider SparseAdam instead")

                state = self.state[p]
                # State initialization
                if len(state) == 0:
                    # Exponential moving average of gradient values
                    state['next_m'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['next_v'] = torch.zeros_like(p.data)

                next_m, next_v = state['next_m'], state['next_v']
                beta1, beta2 = group['betas']

                # Add grad clipping
                if group['max_grad_norm'] > 0:
                    clip_grad_norm_(p, group['max_grad_norm'])

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                next_m.mul_(beta1).add_(1 - beta1, grad)
                next_v.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                update = next_m / (next_v.sqrt() + group['eps'])

                # Just adding the square of the weights to the loss function is
                # *not* # the correct way of using L2 regularization or weight
                # decay with Adam, since that will interact with the m and v
                # parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't
                # interact with the m/v parameters. This is equivalent to adding
                # the square of the weights to the loss with plain
                # (non-momentum) SGD.
                if group['weight_decay'] > 0.0:
                    update += group['weight_decay'] * p.data

                lr = group['lr']
                update_with_lr = lr * update
                p.data.add_(-update_with_lr)

                # No bias correction
                # bias_correction1 = 1 - beta1 ** state['step']
                # bias_correction2 = 1 - beta2 ** state['step']

        return loss



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for maintaining moving average.
"""

from collections import deque
from typing import Deque, Dict, List, Optional, Union, no_type_check

from texar.torch.utils.types import MaybeList, MaybeSeq

__all__ = [
    '_SingleAverageRecorder',
    'AverageRecorder',
]

Scalar = Union[int, float]
ID = Union[int, str]
Record = Union[Dict[ID, Scalar], MaybeSeq[Scalar]]


class _SingleAverageRecorder:
    r"""Maintains the moving average (i.e., the average of the latest N
    records) of a single metric.

    Args:
        size (int, optional): The window size of moving average. If `None`,
            the average of all added records is maintained.
        name (str, optional): name of the recorder. Used when printing.
    """

    def __init__(self, size: Optional[int] = None, name: Optional[str] = None):
        if size is not None and size <= 0:
            raise ValueError("`size` must be > 0 or `None`.")
        self._size = size
        self._q: Deque[Scalar] = deque([])
        self._w: Deque[Scalar] = deque([])
        self._sum = 0.
        self._w_sum: Scalar = 0
        self._name = name

    def add(self, record: Scalar, weight: Optional[Scalar] = None):
        r"""Appends a new record.

        Args:
            record: A scalar; the new record to append.
            weight (optional): A scalar, weight of the new record for
                calculating a weighted average. If `None`, weight is set to
                ``1``.
                For example, :attr:`weight` can be set to batch size and
                :attr:`record` the average value of certain metric on the batch
                in order to calculate the average metric value on a whole
                dataset.

        Returns:
            The (moving) average after appending the record.
        """
        w = weight if weight is not None else 1
        self._w_sum += w
        self._sum += record * w

        if self._size is not None:
            if len(self._q) == self._size:
                w_pop = self._w.popleft()
                self._sum -= self._q.popleft() * w_pop
                self._w_sum -= w_pop
            self._q.append(record)
            self._w.append(w)

        return self.avg()

    def avg(self) -> float:
        r"""Returns the (moving) average.
        """
        if self._w_sum == 0:
            return 0.
        return self._sum / self._w_sum

    def reset(self) -> None:
        r"""Cleans all records.
        """
        self._q.clear()
        self._w.clear()
        self._sum = 0.
        self._w_sum = 0

    def to_str(self, precision: Optional[int] = None) -> str:
        r"""Returns a string of the average value.

        Args:
            precision (int, optional): The number of decimal places to keep in
                the returned string. For example, for an average value of
                ``0.1234``, :python:`precision = 2` leads to ``"0.12"``.

        Returns:
            A string of the average value. If :meth:`name` is given, the
            string is of the format like ``"name: 0.1234"`, otherwise
            the string is of the format like ``"0.1234"``.
        """
        prec_str = "{}"
        if precision is not None:
            prec_str = "{:.%df}" % precision

        avg_str = prec_str.format(self.avg())
        if self._name is not None:
            avg_str = "{}: {}".format(self._name, avg_str)

        return avg_str

    @property
    def name(self) -> str:
        r"""The name of the recorder.
        """
        return self.name


class AverageRecorder:
    r"""Maintains the moving averages (i.e., the average of the latest N
    records) of (possibly multiple) fields.

    Fields are determined by the first call of :meth:`add`.

    Args:
        size (int, optional): The window size of moving average. If `None`,
            the average of all added records is maintained.

    Example:

        .. code-block:: python

            ## Use to maintain moving average of training loss
            avg_rec = AverageRecorder(size=10) # average over latest 10 records
            while training:
                loss_0, loss_1  = ...
                avg_rec.add([loss_0, loss_1])
                # avg_rec.avg() == [0.12343452, 0.567800323]
                # avg_rec.avg(0) == 0.12343452
                # avg_rec.to_str(precision=2, ) == '0.12 0.57'

            ## Use to maintain average of test metrics on the whole test set
            avg_rec = AverageRecorder() # average over ALL records
            while test:
                metric_0, metric_1  = ...
                avg_rec.add({'m0': metric_0, 'm1': metric_1}) # dict is allowed
            print(avg_rec.to_str(precision=4, delimiter=' , '))
            # 'm0: 0.1234 , m1: 0.5678'
            #
            # avg_rec.avg() == {'m0': 0.12343452, 'm1': 0.567800323}
            # avg_rec.avg(0) == 0.12343452

    """
    _recorders: Dict[ID, _SingleAverageRecorder]
    _record_type: type

    def __init__(self, size: Optional[int] = None):
        if size is not None and size <= 0:
            raise ValueError("`size` must be > 0 or `None`.")
        self._size = size
        self._recorders = None  # type: ignore
        self._default_metric_name = "metric"
        self._record_type = None  # type: ignore

    @no_type_check
    def _to_dict(self, record: Record) -> Dict[ID, Scalar]:
        if isinstance(record, dict):
            record_dict = record
        elif isinstance(record, (list, tuple)):
            record_dict = dict(enumerate(record))
        else:
            record_dict = {self._default_metric_name: record}
        return record_dict

    def add(self, record: Record, weight: Optional[Scalar] = None):
        r"""Appends a new record.

        :attr:`record` can be a ``list``, ``dict``, or a single scalar. The
        record type is determined at the first time :meth:`add` is called.
        All subsequent calls to :meth:`add` must have the same type of
        :attr:`record`.

        :attr:`record` in subsequent calls to :meth:`add` can contain only
        a subset of fields than the first call to :meth:`add`.

        Example:

            .. code-block:: python

                recorder.add({'1': 0.2, '2': 0.2}) # 1st call to `add`
                x = recorder.add({'1': 0.4}) # 2nd call to `add`
                # x == {'1': 0.3, '2': 0.2}

        Args:
            record: A single scalar, a list of scalars, or a dict of scalars.
            weight (optional): A scalar, weight of the new record for
                calculating a weighted average. If `None`, weight is set to
                ``1``.
                For example, :attr:`weight` can be set to batch size and
                :attr:`record` the average value of certain metrics on the batch
                in order to calculate the average metric values on a whole
                dataset.

        Returns:
            The (moving) average after appending the record, with the same
            type as :attr:`record`.
        """
        if self._record_type is None:
            self._record_type = type(record)
        elif self._record_type != type(record):
            raise ValueError('The type of `record` is not consistent. '
                             'Expect type `{}`'.format(self._record_type))

        record_dict = self._to_dict(record)
        if self._recorders is None:
            self._recorders = {
                name: _SingleAverageRecorder(
                    self._size, name if self._record_type == dict else None)
                for name in record_dict.keys()
            }

        for name, val in record_dict.items():
            self._recorders[name].add(val, weight=weight)

        return self.avg()

    def avg(self, id_or_name: Optional[MaybeList[ID]] = None) -> Record:
        r"""Returns the (moving) average.

        Args:
            id_or_name (optional): A list of or a single element.
                Each element is the index (if the record type is ``list``) or
                name (if the record type is ``dict``) of the field for which
                the average is calculated. If not given, the average of all
                fields are returned.

        Returns:
            The average value(s). If :attr:`id_or_name` is a single element
            (not a list), then returns the average value of the corresponding
            field. Otherwise, if :attr:`id_or_name` is a list of element(s),
            then returns average value(s) in the same type as :attr:`record`
            of :meth:`add`.
        """
        if self._recorders is None:
            return 0.

        keys = id_or_name
        if keys is None:
            keys = list(self._recorders.keys())

        if not isinstance(keys, (list, tuple)):
            return self._recorders[keys].avg()

        avg = {key: self._recorders[key].avg() for key in keys}
        if self._record_type in {list, tuple}:
            ret_avg = []
            for k, v in avg.items():
                if k in keys:
                    ret_avg.append(v)
            return self._record_type(ret_avg)
        elif self._record_type == dict:
            return avg
        else:
            return avg[self._default_metric_name]

    def reset(self, id_or_name: Optional[MaybeList[ID]] = None):
        r"""Resets the record.

        Args:
            id_or_name (optional): A list or a single element. Each element is
                the index (if the record type is ``list``) or name (if the
                record type is ``dict``) of the field to reset.
                If `None`, all fields are reset.
        """
        keys = id_or_name
        if keys is None:
            keys = list(self._recorders.keys())
        elif not isinstance(keys, (list, tuple)):
            keys = [keys]

        for key in keys:
            self._recorders[key].reset()

    def to_str(self, precision: Optional[int] = None,
               delimiter: str = ' ') -> str:
        r"""Returns a string of the average values of the records.

        Args:
            precision (int, optional): The number of decimal places to keep in
                the returned string. For example, for an average value of
                ``0.1234``, :python:`precision = 2` leads to ``"0.12"``.
            delimiter (str): The delimiter string that separates between
                fields.

        Returns:
            A string of the average values.

            If record is of type ``dict``, the string is a concatenation of
            ``"field_name: average_value"``, delimited with :attr:`delimiter`.
            For example, ``"field_name_1: 0.1234 field_name_2: 0.5678 ..."``.

            Otherwise, the string is of a concatenation of 'average_value'.
            For example, ``"0.1234 0.5678 ..."``
        """
        strs = {name: rec.to_str(precision=precision)
                for name, rec in self._recorders.items()}
        str_list: List[str] = []
        if self._record_type in [list, tuple]:
            # Enumerates the keys in order, which are the indexes
            str_list.extend(strs[i] for i in range(len(strs)))
        elif self._record_type == dict:
            str_list = list(strs.values())
        else:
            str_list = [strs[self._default_metric_name]]

        avg_str = delimiter.join(str_list)

        return avg_str



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various utilities specific to data processing.
"""
import collections
import logging
import os
import sys
import tarfile
import urllib.request
import zipfile
from typing import List, Optional, overload, Union, Dict, Tuple

import numpy as np

from texar.torch.utils import utils_io
from texar.torch.utils.types import MaybeList, MaybeTuple, PathLike

__all__ = [
    "maybe_download",
    "read_words",
    "make_vocab",
    "count_file_lines",
    "get_filename"
]

Py3 = sys.version_info[0] == 3


# TODO: Remove these once pylint supports function stubs.
# pylint: disable=unused-argument,function-redefined,missing-docstring

@overload
def maybe_download(urls: List[str], path: PathLike,
                   filenames: Optional[List[str]] = None,
                   extract: bool = False) -> List[str]: ...


@overload
def maybe_download(urls: str, path: PathLike, filenames: Optional[str] = None,
                   extract: bool = False) -> str: ...


def maybe_download(urls, path, filenames=None, extract=False):
    r"""Downloads a set of files.

    Args:
        urls: A (list of) URLs to download files.
        path (str): The destination path to save the files.
        filenames: A (list of) strings of the file names. If given,
            must have the same length with :attr:`urls`. If `None`,
            filenames are extracted from :attr:`urls`.
        extract (bool): Whether to extract compressed files.

    Returns:
        A list of paths to the downloaded files.
    """
    utils_io.maybe_create_dir(path)

    if not isinstance(urls, (list, tuple)):
        is_list = False
        urls = [urls]
    else:
        is_list = True
    if filenames is not None:
        if not isinstance(filenames, (list, tuple)):
            filenames = [filenames]
        if len(urls) != len(filenames):
            raise ValueError(
                '`filenames` must have the same number of elements as `urls`.')

    result = []
    for i, url in enumerate(urls):
        if filenames is not None:
            filename = filenames[i]
        elif 'drive.google.com' in url:
            filename = _extract_google_drive_file_id(url)
        else:
            filename = url.split('/')[-1]
            # If downloading from GitHub, remove suffix ?raw=True
            # from local filename
            if filename.endswith("?raw=true"):
                filename = filename[:-9]

        filepath = os.path.join(path, filename)
        result.append(filepath)

        # if not tf.gfile.Exists(filepath):
        if not os.path.exists(filepath):
            if 'drive.google.com' in url:
                filepath = _download_from_google_drive(url, filename, path)
            else:
                filepath = _download(url, filename, path)

            if extract:
                logging.info('Extract %s', filepath)
                if tarfile.is_tarfile(filepath):
                    tarfile.open(filepath, 'r').extractall(path)
                elif zipfile.is_zipfile(filepath):
                    with zipfile.ZipFile(filepath) as zfile:
                        zfile.extractall(path)
                else:
                    logging.info("Unknown compression type. Only .tar.gz"
                                 ".tar.bz2, .tar, and .zip are supported")
    if not is_list:
        return result[0]
    return result


# pylint: enable=unused-argument,function-redefined,missing-docstring


def _download(url: str, filename: str, path: str) -> str:
    def _progress_hook(count, block_size, total_size):
        percent = float(count * block_size) / float(total_size) * 100.
        sys.stdout.write(f'\r>> Downloading {filename} {percent:.1f}%')
        sys.stdout.flush()

    filepath = os.path.join(path, filename)
    filepath, _ = urllib.request.urlretrieve(url, filepath, _progress_hook)
    print()
    statinfo = os.stat(filepath)
    print(f'Successfully downloaded {filename} {statinfo.st_size} bytes')

    return filepath


def _extract_google_drive_file_id(url: str) -> str:
    # id is between `/d/` and '/'
    url_suffix = url[url.find('/d/') + 3:]
    if url_suffix.find('/') == -1:
        # if there's no trailing '/'
        return url_suffix
    file_id = url_suffix[:url_suffix.find('/')]
    return file_id


def _download_from_google_drive(url: str, filename: str, path: str) -> str:
    r"""Adapted from `https://github.com/saurabhshri/gdrive-downloader`
    """

    try:
        import requests
    except ImportError:
        print("The requests library must be installed to download files from "
              "Google drive. Please see: https://github.com/psf/requests")
        raise

    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value
        return None

    file_id = _extract_google_drive_file_id(url)

    gurl = "https://docs.google.com/uc?export=download"
    sess = requests.Session()
    response = sess.get(gurl, params={'id': file_id}, stream=True)
    token = _get_confirm_token(response)

    if token:
        params = {'id': file_id, 'confirm': token}
        response = sess.get(gurl, params=params, stream=True)

    filepath = os.path.join(path, filename)
    CHUNK_SIZE = 32768
    with open(filepath, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk:
                f.write(chunk)

    print(f'Successfully downloaded {filename}')

    return filepath


def read_words(filename: str, newline_token: Optional[str] = None) -> List[str]:
    r"""Reads word from a file.

    Args:
        filename (str): Path to the file.
        newline_token (str, optional): The token to replace the original newline
            token "\\n". For example, :python:`tx.data.SpecialTokens.EOS`.
            If `None`, no replacement is performed.

    Returns:
        A list of words.
    """
    with open(filename, "r") as f:
        if Py3:
            if newline_token is None:
                return f.read().split()
            else:
                return f.read().replace("\n", newline_token).split()
        else:
            if newline_token is None:
                return f.read().split()
            else:
                return f.read().replace("\n", newline_token).split()


# TODO: Remove these once pylint supports function stubs.
# pylint: disable=unused-argument,function-redefined,missing-docstring

# A saner overloaded version with default arguments...
@overload
def make_vocab(filenames: MaybeList[str], max_vocab_size: int = -1,
               newline_token: Optional[str] = None) -> List[str]: ...


# ... and an insane version.
@overload
def make_vocab(filenames: MaybeList[str], max_vocab_size: int = -1,
               newline_token: Optional[str] = None,
               return_type: str = "list", return_count: bool = False) \
        -> Union[Union[List[str], Tuple[List[str], List[int]]],
                 MaybeTuple[Dict[str, int]]]: ...


def make_vocab(filenames, max_vocab_size=-1, newline_token=None,
               return_type="list", return_count=False):
    r"""Builds vocab of the files.

    Args:
        filenames (str): A (list of) files.
        max_vocab_size (int): Maximum size of the vocabulary. Low frequency
            words that exceeding the limit will be discarded.
            Set to `-1` (default) if no truncation is wanted.
        newline_token (str, optional): The token to replace the original newline
            token "\\n". For example, :python:`tx.data.SpecialTokens.EOS`.
            If `None`, no replacement is performed.
        return_type (str): Either ``list`` or ``dict``. If ``list`` (default),
            this function returns a list of words sorted by frequency. If
            ``dict``, this function returns a dict mapping words to their index
            sorted by frequency.
        return_count (bool): Whether to return word counts. If `True` and
            :attr:`return_type` is ``dict``, then a count dict is returned,
            which is a mapping from words to their frequency.

    Returns:
        - If :attr:`return_count` is False, returns a list or dict containing
          the vocabulary words.

        - If :attr:`return_count` if True, returns a pair of list or dict
          `(a, b)`, where `a` is a list or dict containing the vocabulary
          words, `b` is a list or dict containing the word counts.
    """

    if not isinstance(filenames, (list, tuple)):
        filenames = [filenames]

    words: List[str] = []
    for fn in filenames:
        words += read_words(fn, newline_token=newline_token)

    counter = collections.Counter(words)
    count_pairs = sorted(counter.items(), key=lambda x: (-x[1], x[0]))

    words, counts = list(zip(*count_pairs))
    words: List[str]
    counts: List[int]
    if max_vocab_size >= 0:
        words = words[:max_vocab_size]
    counts = counts[:max_vocab_size]

    if return_type == "list":
        if not return_count:
            return words
        else:
            return words, counts
    elif return_type == "dict":
        word_to_id = dict(zip(words, range(len(words))))
        if not return_count:
            return word_to_id
        else:
            word_to_count = dict(zip(words, counts))
            return word_to_id, word_to_count
    else:
        raise ValueError(f"Unknown return_type: {return_type}")


# pylint: enable=unused-argument,function-redefined,missing-docstring

def count_file_lines(filenames: MaybeList[str]) -> int:
    r"""Counts the number of lines in the file(s).
    """

    def _count_lines(fn):
        with open(fn, "rb") as f:
            i = -1
            for i, _ in enumerate(f):
                pass
            return i + 1

    if not isinstance(filenames, (list, tuple)):
        filenames = [filenames]
    num_lines = np.sum([_count_lines(fn) for fn in filenames]).item()
    return num_lines


def get_filename(url: str) -> str:
    r"""Extracts the filename of the downloaded checkpoint file from the URL.
    """
    if 'drive.google.com' in url:
        return _extract_google_drive_file_id(url)
    url, filename = os.path.split(url)
    return filename or os.path.basename(url)



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for conv encoders.
"""
import unittest

import torch
from torch import nn

from texar.torch.core.layers import MergeLayer
from texar.torch.modules.encoders.conv_encoders import Conv1DEncoder


class Conv1DEncoderTest(unittest.TestCase):
    r"""Tests :class:`~texar.torch.modules.Conv1DEncoder` class.
    """

    def test_encode(self):
        r"""Tests encode.
        """
        inputs_1 = torch.ones([128, 32, 300])
        encoder_1 = Conv1DEncoder(in_channels=inputs_1.size(1),
                                  in_features=inputs_1.size(2))
        self.assertEqual(len(encoder_1.layers), 4)
        self.assertIsInstance(
            encoder_1.layer_by_name("MergeLayer"), MergeLayer)
        for layer in encoder_1.layers[0].layers:
            self.assertIsInstance(layer, nn.Sequential)

        outputs_1 = encoder_1(inputs_1)
        self.assertEqual(outputs_1.size(), torch.Size([128, 256]))
        self.assertEqual(outputs_1.size(-1), encoder_1.output_size)

        inputs_2 = torch.ones([128, 64, 300])
        hparams = {
            # Conv layers
            "num_conv_layers": 2,
            "out_channels": 128,
            "kernel_size": [[3, 4, 5], 4],
            "other_conv_kwargs": {"padding": 0},
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": 2,
            "pool_stride": 1,
            # Dense layers
            "num_dense_layers": 3,
            "out_features": [128, 128, 10],
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [0, 1, 2],
            "dropout_dense": 2
        }
        network_2 = Conv1DEncoder(in_channels=inputs_2.size(1),
                                  in_features=inputs_2.size(2),
                                  hparams=hparams)
        # dropout-merge-dropout-conv-avgpool-dropout-flatten-
        # (Sequential(Linear,ReLU))-(Sequential(Linear,ReLU))-dropout-linear
        self.assertEqual(len(network_2.layers), 1 + 1 + 1 + 3 + 4 + 1)
        self.assertIsInstance(
            network_2.layer_by_name("MergeLayer"), MergeLayer)
        for layer in network_2.layers[1].layers:
            self.assertIsInstance(layer, nn.Sequential)

        outputs_2 = network_2(inputs_2)
        self.assertEqual(outputs_2.size(), torch.Size([128, 10]))
        self.assertEqual(outputs_2.size(-1), network_2.output_size)


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for conv networks.
"""
import unittest

import torch
from torch import nn

from texar.torch.core.layers import MergeLayer
from texar.torch.modules.networks.conv_networks import Conv1DNetwork


class Conv1DNetworkTest(unittest.TestCase):
    """Tests :class:`~texar.torch.modules.Conv1DNetwork` class.
    """

    def test_feedforward(self):
        """Tests feed forward.
        """
        inputs_1 = torch.ones([128, 32, 300])
        network_1 = Conv1DNetwork(in_channels=inputs_1.shape[1],
                                  in_features=inputs_1.shape[2])
        # dense layers are not constructed yet
        self.assertEqual(len(network_1.layers), 4)
        self.assertIsInstance(
            network_1.layer_by_name("MergeLayer"), MergeLayer)
        for layer in network_1.layers[0].layers:
            self.assertIsInstance(layer, nn.Sequential)

        outputs_1 = network_1(inputs_1)
        self.assertEqual(outputs_1.shape, torch.Size([128, 256]))
        self.assertEqual(outputs_1.size(-1), network_1.output_size)

        inputs_2 = torch.ones([128, 64, 300])
        hparams = {
            # Conv layers
            "num_conv_layers": 2,
            "out_channels": 128,
            "kernel_size": [[3, 4, 5], 4],
            "other_conv_kwargs": {"padding": 0},
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": 2,
            "pool_stride": 1,
            # Dense layers
            "num_dense_layers": 3,
            "out_features": [128, 128, 10],
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [0, 1, 2],
            "dropout_dense": 2
        }

        network_2 = Conv1DNetwork(in_channels=inputs_2.shape[1],
                                  in_features=inputs_2.shape[2],
                                  hparams=hparams)
        # dropout-merge-dropout-(Sequential(Conv, ReLU))-avgpool-dropout-
        # flatten-(Sequential(Linear,ReLU))-(Sequential(Linear,ReLU))-dropout
        # -linear
        self.assertEqual(len(network_2.layers), 1 + 1 + 1 + 3 + 4 + 1)
        self.assertIsInstance(
            network_2.layer_by_name("MergeLayer"), MergeLayer)
        for layer in network_2.layers[1].layers:
            self.assertIsInstance(layer, nn.Sequential)

        outputs_2 = network_2(inputs_2)
        self.assertEqual(outputs_2.shape, torch.Size([128, 10]))
        self.assertEqual(outputs_2.size(-1), network_2.output_size)

        # test whether concatenation happens along channel dim when feature dim
        # has been reduced
        hparams = {
            # Conv layers
            "num_conv_layers": 1,
            "out_channels": 128,
            "kernel_size": [3, 4, 5],
            "other_conv_kwargs": {"padding": 0},
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": None,
            "pool_stride": 1,
            # Dense layers
            "num_dense_layers": 0,
            "out_features": [],
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [],
            "dropout_dense": []
        }

        inputs_3 = torch.ones([128, 64, 300])
        network_3 = Conv1DNetwork(in_channels=inputs_3.shape[1],
                                  in_features=inputs_3.shape[2],
                                  hparams=hparams)
        outputs_3 = network_3(inputs_3)
        num_of_kernels = len(hparams["kernel_size"])
        out_channels = hparams["out_channels"]
        self.assertEqual(outputs_3.shape,
                         torch.Size([128, num_of_kernels * out_channels]))
        self.assertEqual(outputs_3.size(-1), network_3.output_size)

        # test for channels last tensors
        inputs_3 = inputs_3.permute(0, 2, 1)
        outputs_3 = network_3(inputs_3, data_format="channels_last")
        self.assertEqual(outputs_3.shape,
                         torch.Size([128, num_of_kernels * out_channels]))
        self.assertEqual(outputs_3.size(-1), network_3.output_size)

    def test_conv_and_pool_kwargs(self):

        inputs = torch.ones([128, 64, 300])
        hparams = {
            # Conv layers
            "num_conv_layers": 2,
            "out_channels": 128,
            "kernel_size": [[3, 4, 5], 4],
            "other_conv_kwargs": [{"padding": 0}, {"padding": 0}],
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": 2,
            "pool_stride": 1,
            "other_pool_kwargs": {},
            # Dense layers
            "num_dense_layers": 3,
            "out_features": [128, 128, 10],
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [0, 1, 2],
            "dropout_dense": 2
        }

        network = Conv1DNetwork(in_channels=inputs.shape[1],
                                  in_features=inputs.shape[2],
                                  hparams=hparams)
        # dropout-merge-dropout-(Sequential(Conv, ReLU))-avgpool-dropout-
        # flatten-(Sequential(Linear,ReLU))-(Sequential(Linear,ReLU))-dropout
        # -linear
        self.assertEqual(len(network.layers), 1 + 1 + 1 + 3 + 4 + 1)
        self.assertIsInstance(
            network.layer_by_name("MergeLayer"), MergeLayer)
        for layer in network.layers[1].layers:
            self.assertIsInstance(layer, nn.Sequential)

    def test_channel_dimension(self):
        inputs = torch.ones([128, 64, 300])
        hparams = {
            # Conv layers
            "num_conv_layers": 2,
            "out_channels": 128,
            "kernel_size": [[3, 4, 5], 4],
            "other_conv_kwargs": [{"padding": 0}, {"padding": 0}],
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": 2,
            "pool_stride": 1,
            "other_pool_kwargs": {},
            # Dense layers
            "num_dense_layers": 3,
            "out_features": [128, 128, 10],
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [0, 1, 2],
            "dropout_dense": 2
        }

        network = Conv1DNetwork(in_channels=inputs.shape[1],
                                in_features=inputs.shape[2],
                                hparams=hparams)
        # dropout-merge-dropout-(Sequential(Conv, ReLU))-avgpool-dropout-
        # flatten-(Sequential(Linear,ReLU))-(Sequential(Linear,ReLU))-dropout
        # -linear
        self.assertEqual(len(network.layers), 1 + 1 + 1 + 3 + 4 + 1)
        self.assertIsInstance(
            network.layer_by_name("MergeLayer"), MergeLayer)
        for layer in network.layers[1].layers:
            self.assertIsInstance(layer, nn.Sequential)

        inputs_t = torch.ones([128, 300, 64])
        output = network(inputs_t, data_format="channels_last")
        self.assertEqual(output.shape, torch.Size([128, 10]))
        self.assertEqual(output.size(-1), network.output_size)

    def test_sequence_length(self):
        batch_size = 4
        channels = 64
        max_time = 300
        sequence_lengths = torch.LongTensor(batch_size).random_(1, max_time)
        inputs = torch.ones([batch_size, channels, max_time])
        hparams = {
            # Conv layers
            "num_conv_layers": 2,
            "out_channels": 128,
            "kernel_size": [[3, 4, 5], 4],
            "other_conv_kwargs": {"padding": 0},
            # Pooling layers
            "pooling": "AvgPool1d",
            "pool_size": 2,
            "pool_stride": 1,
            # Dense layers
            "num_dense_layers": 0,
            "dense_activation": "ReLU",
            "other_dense_kwargs": None,
            # Dropout
            "dropout_conv": [0, 1, 2],
            "dropout_dense": 2
        }

        network = Conv1DNetwork(in_channels=inputs.shape[1],
                                in_features=inputs.shape[2],
                                hparams=hparams)
        outputs = network(input=inputs, sequence_length=sequence_lengths)
        self.assertEqual(
            outputs.shape,
            torch.Size([batch_size, network.hparams["out_channels"], 884]))

        # channels last
        inputs = torch.ones([batch_size, max_time, channels])
        network = Conv1DNetwork(in_channels=inputs.shape[2],
                                in_features=inputs.shape[1],
                                hparams=hparams)
        outputs = network(input=inputs,
                          sequence_length=sequence_lengths,
                          data_format="channels_last")
        self.assertEqual(
            outputs.shape,
            torch.Size([batch_size, 884, network.hparams["out_channels"]]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Adversarial losses.
"""

from typing import Callable, Tuple

import torch
import torch.nn.functional as F

from texar.torch.utils.types import MaybeTuple

__all__ = [
    'binary_adversarial_losses',
]


def binary_adversarial_losses(
        real_data: torch.Tensor,
        fake_data: torch.Tensor,
        discriminator_fn: Callable[[torch.Tensor], MaybeTuple[torch.Tensor]],
        mode: str = "max_real") -> Tuple[torch.Tensor, torch.Tensor]:
    r"""Computes adversarial losses of real/fake binary discrimination game.

    Example:

    .. code-block:: python

        # Using BERTClassifier as the discriminator, which can accept
        # "soft" token ids for gradient backpropagation
        discriminator = tx.modules.BERTClassifier('bert-base-uncased')

        G_loss, D_loss = tx.losses.binary_adversarial_losses(
            real_data=real_token_ids,  # [batch_size, max_time]
            fake_data=fake_soft_token_ids,  # [batch_size, max_time, vocab_size]
            discriminator_fn=discriminator)

    Args:
        real_data (Tensor or array): Real data of shape
            `[num_real_examples, ...]`.
        fake_data (Tensor or array): Fake data of shape
            `[num_fake_examples, ...]`. `num_real_examples` does not
            necessarily equal `num_fake_examples`.
        discriminator_fn: A callable takes data (e.g., :attr:`real_data` and
            :attr:`fake_data`) and returns the logits of being real. The
            signature of `discriminator_fn` must be:
            :python:`logits, ... = discriminator_fn(data)`.
            The return value of `discriminator_fn` can be the logits, or
            a tuple where the logits are the first element.
        mode (str): Mode of the generator loss. Either "max_real" or "min_fake".

            - **"max_real"** (default): minimizing the generator loss is to
              maximize the probability of fake data being classified as real.

            - **"min_fake"**: minimizing the generator loss is to minimize the
              probability of fake data being classified as fake.

    Returns:
        A tuple `(generator_loss, discriminator_loss)` each of which is
        a scalar Tensor, loss to be minimized.
    """
    real_logits = discriminator_fn(real_data)
    if isinstance(real_logits, (list, tuple)):
        real_logits = real_logits[0]
    real_loss = F.binary_cross_entropy_with_logits(
        real_logits, torch.ones_like(real_logits))

    fake_logits = discriminator_fn(fake_data)
    if isinstance(fake_logits, (list, tuple)):
        fake_logits = fake_logits[0]
    fake_loss = F.binary_cross_entropy_with_logits(
        fake_logits, torch.zeros_like(fake_logits))

    d_loss = real_loss + fake_loss

    if mode == "min_fake":
        g_loss = -fake_loss
    elif mode == "max_real":
        bce_loss = torch.nn.BCEWithLogitsLoss(reduction='mean')
        g_loss = bce_loss(fake_logits, torch.ones_like(fake_logits))
    else:
        raise ValueError("Unknown mode: %s. Only 'min_fake' and 'max_real' "
                         "are allowed.")

    return g_loss, d_loss



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Hyperparameter manager
"""

import copy
import json
from typing import (
    Any, Dict, ItemsView, Iterator, KeysView, Optional, Tuple, Union)

__all__ = [
    'HParams',
]


def _type_name(value):
    return type(value).__name__


class HParams:
    r"""A class that maintains hyperparameters for configuring Texar modules.
    The class has several useful features:

    - **Auto-completion of missing values.** Users can specify only a subset of
      hyperparameters they care about. Other hyperparameters will automatically
      take the default values. The auto-completion performs **recursively** so
      that hyperparameters taking `dict` values will also be auto-completed
      **All Texar modules** provide a :meth:`default_hparams` containing
      allowed hyperparameters and their default values. For example:

        .. code-block:: python

            ## Recursive auto-completion
            default_hparams = {"a": 1, "b": {"c": 2, "d": 3}}
            hparams = {"b": {"c": 22}}
            hparams_ = HParams(hparams, default_hparams)
            hparams_.todict() == {"a": 1, "b": {"c": 22, "d": 3}}
                # "a" and "d" are auto-completed

            ## All Texar modules have built-in `default_hparams`
            hparams = {"dropout_rate": 0.1}
            emb = tx.modules.WordEmbedder(hparams=hparams, ...)
            emb.hparams.todict() == {
                "dropout_rate": 0.1,  # provided value
                "dim": 100            # default value
                ...
            }

    - **Automatic type-check.** For most hyperparameters, provided value must
      have the same or compatible dtype with the default value. :class:`HParams`
      does necessary type-check, and raises Error if improper dtype is provided.
      Also, hyperparameters not listed in `default_hparams` are not allowed,
      except for `"kwargs"` as detailed below.

    - **Flexible dtype for specified hyperparameters.**  Some hyperparameters
      may allow different dtypes of values.

        - Hyperparameters named `"type"` are not type-checked.
          For example, in :func:`~texar.torch.core.get_rnn_cell`, hyperparameter
          `"type"` can take value of an RNNCell class, its string name of module
          path, or an RNNCell class instance. (String name or module path is
          allowed so that users can specify the value in YAML configuration
          files.)

        - For other hyperparameters, list them in the `"@no_typecheck"` field
          in :meth:`default_hparams` to skip type-check. For example, in
          :class:`~texar.torch.modules.Conv1DNetwork`, hyperparameter
          `"kernel_size"` can be set to either a `list` of `int`\ s or simply
          an `int`.

    - **Special flexibility of keyword argument hyperparameters.**
      Hyperparameters named ``"kwargs"`` are used as keyword arguments for a
      class constructor or a function call. Such hyperparameters take a `dict`,
      and users can add arbitrary valid keyword arguments to the dict.
      For example:

        .. code-block:: python

            default_rnn_cell_hparams = {
                "type": "LSTMCell",
                "kwargs": {"num_units": 256}
                # Other hyperparameters
                ...
            }
            my_hparams = {
                "kwargs" {
                    "num_units": 123,
                    # Other valid keyword arguments for LSTMCell constructor
                    "forget_bias": 0.0
                    "activation": "torch.nn.functional.relu"
                }
            }
            _ = HParams(my_hparams, default_rnn_cell_hparams)

    - **Rich interfaces.** An :class:`HParams` instance provides rich interfaces
      for accessing, updating, or adding hyperparameters.

        .. code-block:: python

            hparams = HParams(my_hparams, default_hparams)
            # Access
            hparams.type == hparams["type"]
            # Update
            hparams.type = "GRUCell"
            hparams.kwargs = { "num_units": 100 }
            hparams.kwargs.num_units == 100
            # Add new
            hparams.add_hparam("index", 1)
            hparams.index == 1

            # Convert to `dict` (recursively)
            type(hparams.todic()) == dict

            # I/O
            pickle.dump(hparams, "hparams.dump")
            with open("hparams.dump", 'rb') as f:
                hparams_loaded = pickle.load(f)


    Args:
        hparams: A `dict` or an :class:`HParams` instance containing
            hyperparameters. If `None`, all hyperparameters are set to default
            values.
        default_hparams (dict): Hyperparameters with default values. If `None`,
            Hyperparameters are fully defined by :attr:`hparams`.
        allow_new_hparam (bool): If `False` (default), :attr:`hparams` cannot
            contain hyperparameters that are not included in
            :attr:`default_hparams`, except for the case of :attr:`"kwargs"` as
            above.
    """

    # - The default hyperparameters in :attr:`"kwargs"` are used (for type-check
    #   and complementing missing hyperparameters) only when :attr:`"type"`
    #   takes default value (i.e., missing in :attr:`hparams` or set to
    #   the same value with the default). In this case :attr:`kwargs` allows to
    #   contain new keys not included in :attr:`default_hparams["kwargs"]`.
    #
    # - If :attr:`"type"` is set to an other value and :attr:`"kwargs"` is
    #   missing in :attr:`hparams`, :attr:`"kwargs"` is set to an empty
    #   dictionary.

    def __init__(self, hparams: Optional[Union['HParams', Dict[str, Any]]],
                 default_hparams: Optional[Dict[str, Any]],
                 allow_new_hparam: bool = False):
        if isinstance(hparams, HParams):
            hparams = hparams.todict()
        if default_hparams is not None:
            parsed_hparams = self._parse(
                hparams, default_hparams, allow_new_hparam)
        else:
            parsed_hparams = self._parse(hparams, hparams)
        super().__setattr__('_hparams', parsed_hparams)

    @staticmethod
    def _parse(hparams: Optional[Dict[str, Any]],
               default_hparams: Optional[Dict[str, Any]],
               allow_new_hparam: bool = False):
        r"""Parses hyperparameters.

        Args:
            hparams (dict): Hyperparameters. If `None`, all hyperparameters are
                set to default values.
            default_hparams (dict): Hyperparameters with default values.
                If `None`,Hyperparameters are fully defined by :attr:`hparams`.
            allow_new_hparam (bool): If `False` (default), :attr:`hparams`
                cannot contain hyperparameters that are not included in
                :attr:`default_hparams`, except the case of :attr:`"kwargs"`.

        Return:
            A dictionary of parsed hyperparameters. Returns `None` if both
            :attr:`hparams` and :attr:`default_hparams` are `None`.

        Raises:
            ValueError: If :attr:`hparams` is not `None` and
                :attr:`default_hparams` is `None`.
            ValueError: If :attr:`default_hparams` contains "kwargs" not does
                not contains "type".
        """
        if hparams is None and default_hparams is None:
            return None

        if hparams is None:
            return HParams._parse(default_hparams, default_hparams)

        if default_hparams is None:
            raise ValueError("`default_hparams` cannot be `None` if `hparams` "
                             "is not `None`.")
        no_typecheck_names = default_hparams.get('@no_typecheck', [])

        if "kwargs" in default_hparams and "type" not in default_hparams:
            raise ValueError("Ill-defined hyperparameter structure: 'kwargs' "
                             "must accompany with 'type'.")

        parsed_hparams = copy.deepcopy(default_hparams)

        # Parse recursively for params of type dictionary that are missing
        # in `hparams`.
        for name, value in default_hparams.items():
            if name not in hparams and isinstance(value, dict):
                if (name == 'kwargs' and 'type' in hparams and
                        hparams['type'] != default_hparams['type']):
                    # Set params named "kwargs" to empty dictionary if "type"
                    # takes value other than default.
                    parsed_hparams[name] = HParams({}, {})
                else:
                    parsed_hparams[name] = HParams(value, value)

        # Parse hparams
        for name, value in hparams.items():
            if name not in default_hparams:
                if allow_new_hparam:
                    parsed_hparams[name] = HParams._parse_value(value, name)
                    continue
                raise ValueError(
                    "Unknown hyperparameter: %s. Only hyperparameters "
                    "named 'kwargs' hyperparameters can contain new "
                    "entries undefined in default hyperparameters." % name)

            if value is None:
                parsed_hparams[name] = HParams._parse_value(
                    parsed_hparams[name])

            default_value = default_hparams[name]
            if default_value is None:
                parsed_hparams[name] = HParams._parse_value(value)
                continue

            # Parse recursively for params of type dictionary.
            if isinstance(value, dict):
                if (name not in no_typecheck_names and
                        not isinstance(default_value, dict)):
                    raise ValueError(
                        "Hyperparameter '%s' must have type %s, got %s" %
                        (name, _type_name(default_value), _type_name(value)))
                if name == "kwargs":
                    if ("type" in hparams and
                            hparams["type"] != default_hparams['type']):
                        # Leave "kwargs" as-is if "type" takes value
                        # other than default.
                        parsed_hparams[name] = HParams(value, value)
                    else:
                        # Allow new hyperparameters if "type" takes default
                        # value
                        parsed_hparams[name] = HParams(
                            value, default_value, allow_new_hparam=True)
                elif name in no_typecheck_names:
                    parsed_hparams[name] = HParams(value, value)
                else:
                    parsed_hparams[name] = HParams(
                        value, default_value, allow_new_hparam)
                continue

            # Do not type-check hyperparameter named "type" and accompanied
            # with "kwargs"
            if name == 'type' and 'kwargs' in default_hparams:
                parsed_hparams[name] = value
                continue

            if name in no_typecheck_names:
                parsed_hparams[name] = value
            elif isinstance(value, type(default_value)):
                parsed_hparams[name] = value
            elif callable(value) and callable(default_value):
                parsed_hparams[name] = value
            else:
                try:
                    parsed_hparams[name] = type(default_value)(value)
                except TypeError:
                    raise ValueError(
                        "Hyperparameter '%s' must have type %s, got %s" %
                        (name, _type_name(default_value), _type_name(value)))

        return parsed_hparams

    @staticmethod
    def _parse_value(value: Any, name: Optional[str] = None) -> Any:
        if isinstance(value, dict) and (name is None or name != "kwargs"):
            return HParams(value, None)
        else:
            return value

    def __getattr__(self, name: str) -> Any:
        r"""Retrieves the value of the hyperparameter.
        """
        if name == '_hparams':
            return super().__getattribute__('_hparams')
        if name not in self._hparams:
            # Raise AttributeError to allow copy.deepcopy, etc
            raise AttributeError("Unknown hyperparameter: %s" % name)
        return self._hparams[name]

    def __getitem__(self, name: str) -> Any:
        r"""Retrieves the value of the hyperparameter.
        """
        return self.__getattr__(name)

    def __setattr__(self, name: str, value: Any):
        r"""Sets the value of the hyperparameter.
        """
        if name not in self._hparams:
            raise ValueError(
                "Unknown hyperparameter: %s. Only the `kwargs` "
                "hyperparameters can contain new entries undefined "
                "in default hyperparameters." % name)
        self._hparams[name] = self._parse_value(value, name)

    def items(self) -> ItemsView[str, Any]:
        r"""Returns the list of hyperparameter `(name, value)` pairs.
        """
        return self._hparams.items()

    def keys(self) -> KeysView[str]:
        r"""Returns the list of hyperparameter names.
        """
        return self._hparams.keys()

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        for name, value in self._hparams.items():
            yield name, value

    def __len__(self) -> int:
        return len(self._hparams)

    def __contains__(self, name) -> bool:
        return name in self._hparams

    def __str__(self) -> str:
        r"""Return a string of the hyperparameters.
        """
        hparams_dict = self.todict()
        return json.dumps(hparams_dict, sort_keys=True, indent=2)

    def get(self, name: str, default: Optional[Any] = None) -> Any:
        r"""Returns the hyperparameter value for the given name. If name is not
        available then returns :attr:`default`.

        Args:
            name (str): the name of hyperparameter.
            default: the value to be returned in case name does not exist.
        """
        try:
            return self.__getattr__(name)
        except AttributeError:
            return default

    def add_hparam(self, name: str, value: Any):
        r"""Adds a new hyperparameter.
        """
        if (name in self._hparams) or hasattr(self, name):
            raise ValueError("Hyperparameter name already exists: %s" % name)
        self._hparams[name] = self._parse_value(value, name)

    def todict(self) -> Dict[str, Any]:
        r"""Returns a copy of hyperparameters as a dictionary.
        """
        dict_ = copy.deepcopy(self._hparams)
        for name, value in self._hparams.items():
            if isinstance(value, HParams):
                dict_[name] = value.todict()
        return dict_



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for T5 tokenizer.
"""

import unittest

import os
import tempfile

from texar.torch.utils.test import pretrained_test
from texar.torch.data.tokenizers.t5_tokenizer import T5Tokenizer
from texar.torch.data.data_utils import maybe_download


class T5TokenizerTest(unittest.TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.SAMPLE_VOCAB = maybe_download(
            'https://github.com/google/sentencepiece/blob/master/'
            'python/test/test_model.model?raw=true', self.tmp_dir.name)

        self.tokenizer = T5Tokenizer.load(self.SAMPLE_VOCAB)

        self.tokenizer.save(self.tmp_dir.name)

    def tearDown(self):
        self.tmp_dir.cleanup()

    @pretrained_test
    def test_model_loading(self):
        for pretrained_model_name in T5Tokenizer.available_checkpoints():
            tokenizer = T5Tokenizer(
                pretrained_model_name=pretrained_model_name)

            info = list(os.walk(tokenizer.pretrained_model_dir))
            _, _, files = info[0]

            self.assertIn('sentencepiece.model', files)

            _ = tokenizer.map_text_to_token(u"This is a test")

    def test_roundtrip(self):
        tokenizer = T5Tokenizer.load(self.tmp_dir.name)

        text = 'I saw a girl with a telescope.'
        ids = tokenizer.map_text_to_id(text)
        tokens = tokenizer.map_text_to_token(text)

        self.assertEqual(text, tokenizer.map_id_to_text(ids))
        self.assertEqual(text, tokenizer.map_token_to_text(tokens))

        text = '<extra_id_32> I saw a girl with a telescope.<extra_id_74>'
        ids = tokenizer.map_text_to_id(text)
        tokens = tokenizer.map_text_to_token(text)

        self.assertEqual(text, tokenizer.map_id_to_text(ids))
        self.assertEqual(text, tokenizer.map_token_to_text(tokens))


if __name__ == "__main__":
    unittest.main()



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Various sampler classes.
"""

# pylint: disable=protected-access

from typing import (
    Any, Callable, Generic, Iterator, List, Optional, Tuple, TypeVar, Union)

import torch
from torch.utils.data import sampler as torch_sampler

from texar.torch.data.data.data_base import DatasetBase


__all__ = [
    "BatchingStrategy",
    "TokenCountBatchingStrategy",
]

Example = TypeVar('Example')


# pylint: disable=attribute-defined-outside-init
# TODO: Remove this when Pylint fixes the bug. If the `disable` directive is not
#  added, Pylint incorrectly reports this error for `self.size` in subclasses of
#  `SamplerBase` in Python 3.6 due to use of the Generic class.
#  See Pylint issue: https://github.com/PyCQA/pylint/issues/2981

class SamplerBase(torch_sampler.Sampler, Generic[Example]):
    r"""A subclass of :torch_docs:`~torch.utils.data.Sampler
    <data.html#torch.utils.data.Sampler>` that supports:

    - Returning raw examples when required.
    - Creating iterators with unknown dataset size.

    This class is used internally in
    :class:`~texar.torch.data.data.DataIterator`. It calls the
    :meth:`~texar.torch.data.data.DatasetBase._prefetch_source` method to ensure
    the required number of raw examples are prefetched from source.

    Args:
        data: The :class:`~texar.torch.data.data.DatasetBase` instance.
    """
    size: Optional[int]

    def __init__(self, data: DatasetBase[Any, Example]):
        super().__init__(data)

        self._data = data
        self.size = None

    def _iterator_given_size(self, size: int) -> Iterator[int]:
        r"""Return an iterator that generates samples when the dataset size
        is given.

        Args:
            size: The dataset size.
        """
        raise NotImplementedError

    def _iterator_unknown_size(self) -> Iterator[int]:
        r"""Return an iterator that generates samples when the dataset size
        is unknown. This iterator must also call
        :meth:`texar.torch.data.data.DatasetBase._prefetch_source` and check
        whether the dataset size can be determined, before yielding the index.
        See example implementations for details.
        """
        raise NotImplementedError

    def __iter__(self) -> Union[Iterator[int], Iterator[Tuple[int, Example]]]:
        r"""Return an iterator based on the dataset settings.
        """
        self.size = self._data._dataset_size
        if (not self._data._fully_cached or
                self._data._should_call_prefetch_source):
            self._data._start_iteration()
            # First epoch of lazy loading, calling prefetch, and returning
            # indices and examples.
            iterator = self._iterator_unknown_size()
        else:
            # Non-lazy loading, or when dataset has been fully iterated.
            assert self.size is not None
            iterator = self._iterator_given_size(self.size)

        if self._data._should_call_prefetch_processed:
            # Processing routine is performed in main process. Yield
            # processed examples instead.
            map_fn = lambda idx: (idx, self._data._processed_cache[idx])
        elif self._data._should_yield_raw_example:
            # Return indices and examples for any epoch in this case.
            map_fn = lambda idx: (idx, self._data._source[idx])
        else:
            map_fn = None  # type: ignore
        if map_fn is not None:
            return map(map_fn, iterator)

        return iterator

    def __len__(self):
        if self.size is not None:
            return self.size
        raise AttributeError("Dataset size cannot be determined at this point")


class SequentialSampler(SamplerBase[Example]):
    r"""Samples elements sequentially, always in the same order. Same as
    :torch_docs:`~torch.utils.data.SequentialSampler
    <data.html#torch.utils.data.SequentialSampler>`
    """

    def _iterator_given_size(self, size: int) -> Iterator[int]:
        return iter(range(size))

    def _iterator_unknown_size(self) -> Iterator[int]:
        index = 0
        while True:
            cur_size = self._data._prefetch_source(index)
            if cur_size is not None:
                self.size = cur_size
                break
            yield index
            index += 1


class RandomSampler(SamplerBase[Example]):
    r"""Samples elements randomly. If without replacement, then sample from a
    shuffled dataset. If with replacement, then user can specify ``num_samples``
    to draw.

    This class uses :torch_docs:`torch.utils.data.RandomSampler
    <data.html#torch.utils.data.RandomSampler>` directly. Given the
    nature of such shuffling, it cannot be used for iterators with unknown size.

    Args:
        data: The :class:`~texar.torch.data.data.DatasetBase` instance.
        num_samples (int): number of samples to draw, default=len(dataset)
        replacement (bool): samples are drawn with replacement if `True`,
            default=False
    """

    def __init__(self, data: DatasetBase[Any, Example],
                 replacement: bool = False, num_samples: Optional[int] = None):
        super().__init__(data)
        self._sampler = torch_sampler.RandomSampler(
            data, replacement, num_samples)

    def _iterator_given_size(self, size: int) -> Iterator[int]:
        del size  # not used
        return iter(self._sampler)

    def _iterator_unknown_size(self) -> Iterator[int]:
        raise TypeError(
            "RandomSampler does not support lazy data loading. To perform "
            "shuffling with lazy loading, use BufferShuffleSampler.")


class BufferShuffleSampler(SamplerBase[Example]):
    r"""A :torch_docs:`~torch.utils.data.Sampler
    <data.html#torch.utils.data.Sampler>` that uses a shuffle buffer, as
    in TensorFlow. The buffer is first filled with data examples. Each time a
    sample is drawn from the buffer, and the drawn sample is replaced with the
    next data example.

    This class is used internally in
    :class:`~texar.torch.data.data.DataIterator`. It calls the
    :meth:`~texar.torch.data.data.DatasetBase._prefetch_source` method to ensure
    the required number of raw examples are prefetched from source.

    Args:
        data: The :class:`~texar.torch.data.data.DatasetBase` instance.
        buffer_size: The size of the shuffle buffer. Use larger buffer sizes for
            more uniformly-random shuffling.
    """

    def __init__(self, data: DatasetBase[Any, Example], buffer_size: int):
        super().__init__(data)
        self.buffer_size = buffer_size

    def _iterator_given_size(self, size) -> Iterator[int]:
        if self.buffer_size >= size:
            yield from iter(torch.randperm(size).tolist())
            return

        buffer = list(range(self.buffer_size))
        for x in range(self.buffer_size, size):
            sample = torch.randint(self.buffer_size, (1,)).item()
            index = buffer[sample]
            yield index
            buffer[sample] = x
        yield from (buffer[x] for x in torch.randperm(self.buffer_size))

    def _iterator_unknown_size(self) -> Iterator[int]:
        buffer = list(range(self.buffer_size))
        x = self.buffer_size
        while True:
            sample = torch.randint(self.buffer_size, (1,)).item()
            index = buffer[sample]
            cur_size = self._data._prefetch_source(index)
            if cur_size is not None and index >= cur_size:
                self.size = cur_size
            if self.size is not None and index >= self.size:
                break
            yield index
            buffer[sample] = x
            x += 1
        yield from (buffer[x] for x in torch.randperm(self.buffer_size)
                    if buffer[x] < self.size)


# pylint: enable=attribute-defined-outside-init


class BatchingStrategy(Generic[Example]):
    r"""Decides batch boundaries in dynamic batching. Please refer to
    :class:`TokenCountBatchingStrategy` for a concrete example.
    """

    def reset_batch(self) -> None:
        r"""Reset the internal state of the batching strategy. This method is
        called at the start of iteration, and after each batch is yielded.
        """
        raise NotImplementedError

    def add_example(self, example: Example) -> bool:
        r"""Add an example into the current batch, and modify internal states
        accordingly. If the example should not be added to the batch, this
        method does not modify the internal state, and returns `False`.

        Args:
            example: The example to add to the batch.

        Returns:
            A boolean value indicating whether :attr:`example` should be added
            to the batch.
        """
        raise NotImplementedError


class TokenCountBatchingStrategy(BatchingStrategy[Example]):
    r"""Create dynamically-sized batches so that the total number of tokens
    inside each batch is constrained.

    Args:
        max_tokens (int): The maximum number of tokens inside each batch.
        max_batch_size (int, optional): The maximum number of examples for each
            batch. If `None`, batches can contain arbitrary number of examples
            as long as the total number of tokens does not exceed
            :attr:`max_tokens`.
        length_fn (callable, optional): A function taking a data example as
            argument, and returning the number of tokens in the example. By
            default, :python:`len` is used, which is the desired behavior if the
            dataset in question is a :class:`~texar.torch.data.MonoTextData`.
    """
    sum_tokens: int
    cur_batch_size: int

    def __init__(self, max_tokens: int, max_batch_size: Optional[int] = None,
                 length_fn: Optional[Callable[[Example], int]] = None):
        self.max_batch_size = max_batch_size
        self.max_tokens = max_tokens
        self.length_fn: Callable[[Example], int]
        self.length_fn = length_fn or len  # type: ignore

    def reset_batch(self) -> None:
        self.sum_tokens = 0
        self.cur_batch_size = 0

    def add_example(self, example: Example) -> bool:
        if self.cur_batch_size == self.max_batch_size:
            return False
        cur_tokens = self.length_fn(example)
        if cur_tokens + self.sum_tokens > self.max_tokens:
            return False

        self.cur_batch_size += 1
        self.sum_tokens += cur_tokens
        return True


class DynamicBatchSampler(torch_sampler.BatchSampler, Generic[Example]):
    r"""A subclass of :torch_docs:`~torch.utils.data.BatchSampler
    <data.html#torch.utils.data.BatchSampler>` that supports dynamic batching
    through a user-provided :class:`BatchingStrategy`. This class is used
    internally.

    Args:
        dataset: The dataset to create batches from.
        sampler: An instance of :class:`SamplerBase` that returns indices of
            each sampled example.
        strategy: An instance of :class:`BatchingStrategy` that decides whether
            a batch should be yielded.
    """

    def __init__(self, dataset: DatasetBase[Any, Example],  # pylint: disable=super-init-not-called
                 sampler: SamplerBase, strategy: BatchingStrategy[Example]):
        self.dataset = dataset
        self.sampler = sampler
        self.strategy = strategy

    def __iter__(self) -> Union[Iterator[List[int]],  # type: ignore
                                Iterator[List[Tuple[int, Example]]]]:
        batch = []  # type: ignore
        self.strategy.reset_batch()
        for idx in self.sampler:
            if isinstance(idx, tuple):
                example = self.dataset[idx[0]]
            else:
                example = self.dataset[idx]
            while not self.strategy.add_example(example):
                if len(batch) == 0:
                    raise ValueError(f"Batching strategy refused to add "
                                     f"example {idx} to empty batch.")
                yield batch
                batch = []
                self.strategy.reset_batch()
            batch.append(idx)
        if len(batch) > 0:
            yield batch
            self.strategy.reset_batch()

    def __len__(self):
        raise TypeError("DynamicBatchSampler does not support __len__")



# Copyright 2019 The Texar Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
SentencePiece Tokenizer.
"""

from typing import Any, Dict, List, Optional, Tuple

import os
from shutil import copyfile, move

import sentencepiece as spm

from texar.torch.data.tokenizers.tokenizer_base import TokenizerBase
from texar.torch.modules.pretrained.pretrained_base import default_download_dir
from texar.torch.utils.utils_io import maybe_create_dir

__all__ = [
    "SentencePieceTokenizer"
]


class SentencePieceTokenizer(TokenizerBase):
    r"""SentencePiece Tokenizer. This class is a wrapper of Google's
    `SentencePiece`_ with richer ready-to-use functionalities such as
    adding tokens and saving/loading.

    `SentencePiece` is an unsupervised text tokenizer mainly for Neural
    Network-based text generation systems where the vocabulary size
    is predetermined prior to the neural model training. `SentencePiece`
    implements sub-word units (e.g., byte-pair-encoding (BPE) and unigram
    language model) with the extension of direct training from raw sentences.

    The supported algorithms in `SentencePiece` are: ``bpe``, ``word``,
    ``char``, and ``unigram``, which is specified in :attr:`hparams`.

    Args:
        cache_dir (optional): the path to a folder in which the
            trained `sentencepiece` model will be cached. If `None` (default),
            a default directory (``texar_data`` folder under user's home
            directory) will be used.
        hparams (dict or HParams, optional): Hyperparameters. Missing
            hyperparameter will be set to default values. See
            :meth:`default_hparams` for the hyperparameter structure
            and default values.

    .. _`SentencePiece`: https://github.com/google/sentencepiece
    """

    _IS_PRETRAINED = False
    _VOCAB_FILE_NAMES = {
        'vocab_file': 'spiece.model',
    }
    _TRAIN_ARG_MAP = {
        'text_file': 'input',
        'model_type': 'model_type',
        'vocab_size': 'vocab_size',
        'bos_token': 'bos_piece',
        'eos_token': 'eos_piece',
        'unk_token': 'unk_piece',
        'pad_token': 'pad_piece',
    }

    def __init__(self,
                 cache_dir: Optional[str] = None,
                 hparams=None):
        super().__init__(hparams=hparams)

        self.__dict__: Dict

        if self.hparams['vocab_file'] is not None:
            self.vocab_file = self.hparams['vocab_file']
            self.sp_model = spm.SentencePieceProcessor()
            self.sp_model.Load(self.vocab_file)

            bos_id = self.sp_model.bos_id()
            eos_id = self.sp_model.eos_id()
            unk_id = self.sp_model.unk_id()
            pad_id = self.sp_model.pad_id()

            self.bos_token = None
            if bos_id != -1:
                self.bos_token = self.sp_model.IdToPiece(bos_id)

            self.eos_token = None
            if eos_id != -1:
                self.eos_token = self.sp_model.IdToPiece(eos_id)

            self.unk_token = None
            if unk_id != -1:
                self.unk_token = self.sp_model.IdToPiece(unk_id)

            self.pad_token = None
            if pad_id != -1:
                self.pad_token = self.sp_model.IdToPiece(pad_id)

        elif self.hparams['text_file'] is not None:
            cmd = ['--model_prefix=spiece']
            for arg, val in self.hparams.items():
                if arg in self._TRAIN_ARG_MAP:
                    if arg in self._SPECIAL_TOKENS_ATTRIBUTES and val is None:
                        cmd.append('--' + arg.replace('token', 'id') + '=-1')
                    else:
                        cmd.append('--' + self._TRAIN_ARG_MAP[arg] + '=' +
                                   str(val))

            cache_path = self.train(" ".join(cmd), cache_dir)
            self.vocab_file = os.path.join(
                cache_path, self._VOCAB_FILE_NAMES['vocab_file'])
            self.sp_model = spm.SentencePieceProcessor()
            self.sp_model.Load(self.vocab_file)
        else:
            raise ValueError("'vocab_file' and 'text_file' can not be None "
                             "at the same time.")

    @classmethod
    def train(cls, cmd: str,  # type: ignore
              cache_dir: Optional[str] = None) -> str:
        r"""Trains the tokenizer from the raw text file. This function is
        a wrapper of `sentencepiece.SentencePieceTrainer.Train`_ function.

        Example:

        .. code-block:: python

            SentencePieceTokenizer.train('--input=test/botchan.txt
            --model_prefix=m --vocab_size=1000')

        Args:
            cmd (str): the command for the tokenizer training procedure.
                See ``sentencepiece.SentencePieceTrainer.Train`` for the
                detailed usage.
            cache_dir (optional): the path to a folder in which the trained
                `sentencepiece` model will be cached. If `None` (default),
                a default directory (`texar_pytorch` folder under user's home
                directory) will be used.

        Returns:
            Path to the cache directory.

        .. _`sentencepiece.SentencePieceTrainer.Train`:
            https://github.com/google/sentencepiece/blob/master/python/sentencepiece.py
        """
        if cache_dir is None:
            cache_path = str(default_download_dir('SentencePiece'))
        else:
            if not os.path.isdir(cache_dir):
                raise ValueError(f"Cache directory ({cache_dir}) should be a "
                                 f"directory.")
            cache_path = os.path.abspath(cache_dir)

        maybe_create_dir(cache_path)

        spm.SentencePieceTrainer.Train(cmd)
        cwd = os.getcwd()

        vocab_file = os.path.join(cwd, cls._VOCAB_FILE_NAMES['vocab_file'])
        out_vocab_file = os.path.join(
            cache_path, cls._VOCAB_FILE_NAMES['vocab_file'])

        if os.path.abspath(vocab_file) != os.path.abspath(out_vocab_file):
            move(vocab_file, out_vocab_file)

        # Delete spiece.vocab (We might want to keep it as well)
        extra_file = vocab_file.rstrip('model') + 'vocab'
        os.remove(extra_file)

        return cache_path

    # spm.SentencePieceProcessor() is a SwigPyObject object which cannot be
    # pickled. We need to define __getstate__ here.
    def __getstate__(self):
        state = self.__dict__.copy()
        state["sp_model"] = None
        state["vocab_file"] = None
        return state, self.vocab_file

    # spm.SentencePieceProcessor() is a SwigPyObject object which cannot be
    # pickled. We need to define __setstate__ here.
    def __setstate__(self, d):
        self.__dict__, self.vocab_file = d
        self.sp_model = spm.SentencePieceProcessor()
        self.sp_model.Load(self.vocab_file)

    def save_vocab(self, save_dir: str) -> Tuple[str]:
        r"""Save the sentencepiece vocabulary (copy original file) to
        a directory.
        """
        if not os.path.isdir(save_dir):
            raise ValueError("Vocabulary path ({}) should be a "
                             "directory".format(save_dir))
        out_vocab_file = os.path.join(
            save_dir, self._VOCAB_FILE_NAMES['vocab_file'])

        if os.path.abspath(self.vocab_file) != os.path.abspath(out_vocab_file):
            copyfile(self.vocab_file, out_vocab_file)

        return (out_vocab_file,)

    @property
    def vocab_size(self) -> int:
        return len(self.sp_model)

    def _map_text_to_token(self, text: str) -> List[str]:  # type: ignore
        return self.sp_model.EncodeAsPieces(text)

    def _map_token_to_id(self, token: str) -> int:
        return self.sp_model.PieceToId(token)

    def _map_id_to_token(self, index: int) -> str:
        token = self.sp_model.IdToPiece(index)
        return token

    def map_token_to_text(self, tokens: List[str]) -> str:
        r"""Maps a sequence of tokens (string) in a single string."""
        out_string = self.sp_model.DecodePieces(tokens)
        return out_string

    @staticmethod
    def default_hparams() -> Dict[str, Any]:
        r"""Returns a dictionary of hyperparameters with default values.

        * If `hparams['vocab_file']` is specified, the tokenizer is directly
          loaded from the vocabulary file. In this case, all other
          configurations in `hparams` are ignored.
        * Otherwise, the tokenizer is automatically trained based on
          `hparams['text_file']`. In this case, `hparams['vocab_size']` must
          be specified.
        * `hparams['vocab_file']` and `hparams['text_file']` can not be None
          at the same time.

        .. code-block:: python

            {
                "vocab_file": None,
                "text_file": None,
                "vocab_size": None,
                "model_type": "unigram",
                "bos_token": "<s>",
                "eos_token": "</s>",
                "unk_token": "<unk>",
                "pad_token": "<pad>",
            }

        Here:

        `"vocab_file"`: str or None
            The path to a sentencepiece vocabulary file.

        `"text_file"`: str or None
            Comma separated list of input sentences.

        `"vocab_size"`: int or None
            Vocabulary size. The user can specify the vocabulary size, and the
            tokenizer training procedure will train and yield a vocabulary
            of the specified size.

        `"model_type"`: str
            Model algorithm to train the tokenizer. Available algorithms are:
            ``bpe``, ``word``, ``char``, and ``unigram``.

        `"bos_token"`: str or None
            Beginning of sentence token. Set None to disable ``bos_token``.

        `"eos_token"`: str or None
            End of sentence token. Set None to disable ``eos_token``.

        `"unk_token"`: str or None
            Unknown token. Set None to disable ``unk_token``.

        `"pad_token"`: str or None
            Padding token. Set None to disable ``pad_token``.
        """
        return {
            'vocab_file': None,
            'text_file': None,
            'vocab_size': None,
            'model_type': 'unigram',
            'bos_token': '<s>',
            'eos_token': '</s>',
            'pad_token': '<pad>',
            'unk_token': '<unk>',
        }

    @classmethod
    def _transform_config(cls, pretrained_model_name: str,
                          cache_dir: str) -> Dict[str, Any]:
        pass

    def _init_from_checkpoint(self, pretrained_model_name: str,
                              cache_dir: str, **kwargs):
        pass
